# -*- coding: utf-8 -*-
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

from .layers.ff import FF
from .layers.attention import Attention


def get_rnn_hidden_state(h):
    """Returns h_t transparently regardless of RNN type."""
    return h if not isinstance(h, tuple) else h[0]


class ConditionalDecoder(nn.Module):
    """A conditional decoder with attention à la dl4mt-tutorial."""

    def __init__(self, input_size, hidden_size, n_vocab, rnn_type,
                 encoder_size=200, tied_emb=False,
                 dec_init='zero', dec_init_activ='tanh', dec_init_size=None,
                 att_type='mlp', att_activ='tanh', att_bottleneck='ctx', att_temp=1.0,
                 transform_ctx=True, mlp_bias=False, dropout_out=0, **kwargs):
        super().__init__()

        # Normalize case
        self.rnn_type = rnn_type.upper()

        # Safety checks
        assert self.rnn_type in ('GRU', 'LSTM'), \
            "rnn_type '{}' not known".format(rnn_type)
        assert dec_init in ('zero', 'mean_ctx', 'feats'), \
            "dec_init '{}' not known".format(dec_init)

        RNN = getattr(nn, '{}Cell'.format(self.rnn_type))
        # LSTMs have also the cell state
        self.n_states = 1 if self.rnn_type == 'GRU' else 2

        # Set custom handlers for GRU/LSTM
        if self.rnn_type == 'GRU':
            self._rnn_unpack_states = lambda x: x
            self._rnn_pack_states = lambda x: x
        elif self.rnn_type == 'LSTM':
            self._rnn_unpack_states = self._lstm_unpack_states
            self._rnn_pack_states = self._lstm_pack_states

        # Other arguments
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.encoder_size = encoder_size
        self.n_vocab = n_vocab
        self.tied_emb = tied_emb
        self.dec_init = dec_init
        self.dec_init_size = dec_init_size
        self.dec_init_activ = dec_init_activ
        self.att_type = att_type
        self.att_bottleneck = att_bottleneck
        self.att_activ = att_activ
        self.att_temp = att_temp
        self.transform_ctx = transform_ctx
        self.mlp_bias = mlp_bias
        self.dropout_out = dropout_out

        # Create target embeddings
        self.emb = nn.Embedding(self.n_vocab, self.input_size,
                                padding_idx=0)

        # Create attention layer
        self.att = Attention(self.encoder_size, self.hidden_size,
                             transform_ctx=self.transform_ctx,
                             mlp_bias=self.mlp_bias,
                             att_type=self.att_type,
                             att_activ=self.att_activ,
                             att_bottleneck=self.att_bottleneck,
                             temp=self.att_temp)

        # Decoder initializer FF (for 'mean_ctx' or auxiliary 'feats')
        if self.dec_init in ('mean_ctx', 'feats'):
            if self.dec_init == 'mean_ctx':
                self.dec_init_size = self.encoder_size
            self.ff_dec_init = FF(
                self.dec_init_size,
                self.hidden_size * self.n_states, activ=self.dec_init_activ)

        # Create first decoder layer necessary for attention
        self.dec0 = RNN(self.input_size, self.hidden_size)
        self.dec1 = RNN(self.hidden_size, self.hidden_size)

        # Output dropout
        if self.dropout_out > 0:
            self.do_out = nn.Dropout(p=self.dropout_out)

        # Output bottleneck: maps hidden states to target emb dim
        self.hid2out = FF(self.hidden_size, self.input_size,
                          bias_zero=True, activ='tanh')

        # Final softmax
        self.out2prob = FF(self.input_size, self.n_vocab)

        # Tie input embedding matrix and output embedding matrix
        if self.tied_emb:
            self.out2prob.weight = self.emb.weight

        self.nll_loss = nn.NLLLoss(reduction="sum", ignore_index=0)

    def _lstm_pack_states(self, h):
        return torch.cat(h, dim=-1)

    def _lstm_unpack_states(self, h):
        # Split h_t and c_t into two tensors and return a tuple
        return torch.split(h, self.hidden_size, dim=-1)

    def f_init(self, ctx_dict):
        """Returns the initial h_0 for the decoder."""
        if self.dec_init == 'zero':
            ctx, _ = ctx_dict['enc']
            h_0 = torch.zeros(ctx.shape[1], self.hidden_size * self.n_states)
            return Variable(h_0).cuda()
        elif self.dec_init == 'mean_ctx':
            ctx, ctx_mask = ctx_dict['enc']
            if ctx_mask is None:
                return self.ff_dec_init(ctx.mean(0))
            else:
                return self.ff_dec_init(ctx.sum(0) / ctx_mask.sum(0).unsqueeze(1))
        elif self.dec_init == 'feats':
            ctx, _ = ctx_dict['feats']
            return self.ff_dec_init(ctx.squeeze(0))
        else:
            raise NotImplementedError(self.dec_init)

    def f_next(self, ctx_dict, y, h):
        # Get hidden states from the first decoder (purely cond. on LM)
        h1_c1 = self.dec0(y, self._rnn_unpack_states(h))
        h1 = get_rnn_hidden_state(h1_c1)

        # Apply attention
        self.txt_alpha_t, txt_z_t = self.att(
            h1.unsqueeze(0), *ctx_dict['enc'])

        # Run second decoder (h1 is compatible now as it was returned by GRU)
        h2_c2 = self.dec1(txt_z_t, h1_c1)
        h2 = get_rnn_hidden_state(h2_c2)

        # This is a bottleneck to avoid going from H to V directly
        logit = self.hid2out(h2)

        # Apply dropout if any
        if self.dropout_out > 0:
            logit = self.do_out(logit)

        # Transform logit to T*B*V (V: vocab_size)
        # Compute log_softmax over token dim
        log_p = F.log_softmax(self.out2prob(logit), dim=-1)

        # Return log probs and new hidden states
        return log_p, self._rnn_pack_states(h2_c2)

    def forward(self, ctx_dict, y):
        """Computes the softmax outputs given source annotations `ctxs` and
        ground-truth target token indices `y`. Only called during training.

        Arguments:
            ctxs(Variable): A variable of `S*B*ctx_dim` representing the source
                annotations in an order compatible with ground-truth targets.
            y(Variable): A variable of `T*B` containing ground-truth target
                token indices for the given batch.
        """
        self.alphas = []

        loss = 0.0
        logps = None if self.training else torch.zeros(
            y.shape[0] - 1, y.shape[1], self.n_vocab).cuda()

        # Convert token indices to embeddings -> T*B*E
        y_emb = self.emb(y)

        # Get initial hidden state
        h = self.f_init(ctx_dict)

        # -1: So that we skip the timestep where input is <eos>
        for t in range(y_emb.shape[0] - 1):
            log_p, h = self.f_next(ctx_dict, y_emb[t], h)
            if not self.training:
                logps[t] = log_p.data
            loss += self.nll_loss(log_p, y[t + 1])

        return {'loss': loss, 'logps': logps}



# -*- coding: utf-8 -*-
import os
import shutil
import pathlib
import subprocess


def listify(l):
    """Encapsulate l with list[] if not."""
    return [l] if not isinstance(l, list) else l


class METEORScorer:
    def __init__(self):
        self.jar = os.path.join(os.path.dirname(__file__), "meteor-1.5.jar")
        self.__cmdline = ["java", "-Xmx2G", "-jar", self.jar,
                          "-", "-", "-stdio"]
        self.env = os.environ
        self.env['LC_ALL'] = 'en_US.UTF-8'

        # Sanity check
        if shutil.which('java') is None:
            raise RuntimeError('METEOR requires java which is not installed.')

    def compute(self, refs, hyps, language="en"):
        cmdline = self.__cmdline[:]
        refs = listify(refs)

        if isinstance(hyps, str):
            # If file, open it for line reading
            hyps = open(hyps)

        if language == "auto":
            # Take the extension of the 1st reference file, e.g. ".de"
            language = pathlib.Path(refs[0]).suffix[1:]

        cmdline.extend(["-l", language])

        # Make reference files a list
        iters = [open(f) for f in refs]
        iters.append(hyps)

        # Run METEOR process
        proc = subprocess.Popen(cmdline,
                                stdout=subprocess.PIPE,
                                stdin=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=self.env,
                                universal_newlines=True, bufsize=1)

        eval_line = 'EVAL'

        for line_ctr, lines in enumerate(zip(*iters)):
            lines = [l.rstrip('\n') for l in lines]
            refstr = " ||| ".join(lines[:-1])
            line = "SCORE ||| " + refstr + " ||| " + lines[-1]

            proc.stdin.write(line + '\n')
            eval_line += ' ||| {}'.format(proc.stdout.readline().strip())

        # Send EVAL line to METEOR
        proc.stdin.write(eval_line + '\n')

        # Dummy read segment scores
        for i in range(line_ctr + 1):
            proc.stdout.readline().strip()

        # Compute final METEOR
        try:
            score = float(proc.stdout.readline().strip())
        except Exception as e:
            score = 0.0
        finally:
            # Close METEOR process
            proc.stdin.close()
            proc.terminate()
            proc.kill()
            proc.wait(timeout=2)
            return score



import logging
import os


class LoggingFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    green = "\x1b[33;92m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: green + format + reset,
        logging.INFO: format,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def set_logger(ckpt_dir, seed):
    ch = logging.StreamHandler()
    ch.setFormatter(LoggingFormatter())
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)",
        handlers=[
            logging.FileHandler(os.path.join(ckpt_dir, "{}.log".format(seed)), mode='a+'),
            ch
        ]
    )
    addLoggingLevel("SETTINGS", levelNum=logging.DEBUG)

def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
       raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
       raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
       raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)
    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)



from .NLG.rouge import ROUGEScorer
from .NLG.bleu.bleu import BLEUScorer
from .NLG.meteor.meteor import METEORScorer
from .NLG.cider.cider import Cider
from .NLG.ciderd.ciderd import CiderD
from .CheXbert.chexbert import CheXbert


import torch
from torch.utils.data import Dataset
from .base.ImageDataset import ImageDataset
from .base.TextDataset import TextDataset
import json


class ImSeq(Dataset):
    def __init__(self, seq, image, split, ckpt_dir, **kwargs):
        self.split = split
        self.seq = TextDataset(**seq, split=split, ckpt_dir=ckpt_dir)
        self.image = ImageDataset(**image, split=split)

        assert len(self.image) == len(self.seq)

        # For decoding, if needed
        self.tokenizer = self.seq.tokenizer
        self.tokenizer_max_len = self.seq.tokenizer_max_len

        # For tokenizing
        self.tokenizer_args = self.seq.tokenizer_args

    def __getitem__(self, index):
        return {**self.image.__getitem__(index), **self.seq.__getitem__(index)}

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {**self.seq.get_collate_fn()(batch), **self.image.get_collate_fn()(batch)}
            return collated

        return collate_fn

    def __len__(self):
        return len(self.image)

    def __repr__(self):
        return "ImSeq\n" + str(self.seq) + '\n' + str(self.image)

    def inference(self, seq=None, image=None):
        if seq is None and image is None:
            return dict()

        batch = {}
        if image is not None:
            batch.update(self.image.inference(image))

        if seq is not None:
            batch.update(self.seq.inference(seq))

        assert len(set([len(v) for k, v in batch.items()])) == 1, 'elements in batch do not have the same size'
        return batch



import os
import tqdm
import json
import sys

from torch.utils.data import Dataset
from transformers import AutoTokenizer
from transformers import BertTokenizer
from .utils import Vocab, load_file
from .papers.report_preprocessing import *

import seaborn as sns
import matplotlib.pyplot as plt

sns.set_theme(style="darkgrid")
os.environ["TOKENIZERS_PARALLELISM"] = "false"  # https://github.com/huggingface/transformers/issues/5486


def split_sentences(sentences, processing):
    return [processing(s.strip()).split() for s in sentences]


def make_sentences(root, split, file, processing):
    sentences = load_file(os.path.join(root, split + '.' + file))
    return split_sentences(sentences, processing)


class TextDataset(Dataset):
    def __init__(self,
                 root=None,
                 file=None,
                 split=None,
                 ckpt_dir=None,
                 processing=None,
                 tokenizer=None,
                 tokenizer_max_len=None,
                 vocab_file=None,
                 source='src',
                 show_length=False,
                 **kwargs):

        assert source in ["src", "tgt"]
        assert split is not None, "Argument split cannot be None"
        assert not (file is not None and vocab_file is not None), "You cannot mention both a data file and a vocab file"
        assert not (
                vocab_file is not None and tokenizer is not None), "You cannot mention both a pretrained tokenizer and a vocab file"
        assert not (source == "tgt" and tokenizer_max_len is None), "You must specify tokenizer_max_len for source tgt"

        self.root = root
        self.file = file
        self.split = split
        self.source = source
        self.ckpt_dir = ckpt_dir
        self.processing = eval(processing or 'lambda x: x')
        self.tokenizer_max_len = tokenizer_max_len
        self.vocab_file = vocab_file
        self.sentences = None

        if file is not None:
            self.sentences = make_sentences(root, split, file, self.processing)

        # Create tokenizer from pretrained or vocabulary file
        if tokenizer is not None:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        else:
            if vocab_file is None:
                vocab_file = os.path.join(ckpt_dir, 'vocab.{}'.format(source))
                if split == 'train':
                    vocab = Vocab(self.sentences)
                    if not os.path.exists(vocab_file):
                        vocab.dump(vocab_file)
            self.tokenizer = BertTokenizer(vocab_file=vocab_file,
                                           do_basic_tokenize=False)

        # Create tokenizer forwards args
        self.tokenizer_args = {'return_tensors': 'pt', 'padding': True, 'add_special_tokens': True}
        if self.source == 'src':
            self.tokenizer_args.update({'add_special_tokens': False})
        if self.tokenizer_max_len is not None:
            self.tokenizer_args.update({'padding': 'max_length',
                                        'truncation': True,
                                        'max_length': self.tokenizer_max_len})

        if show_length:
            print(self)
            self.show_length()
            sys.exit()

    def __getitem__(self, index):
        return {'{}_seq'.format(self.source): ' '.join(self.sentences[index])}

    def get_collate_fn(self):
        def collate_fn(batch):
            seq = self.tokenizer([s['{}_seq'.format(self.source)] for s in batch], **self.tokenizer_args)
            collated = {
                'input_ids': seq.input_ids,
                'attention_mask': seq.attention_mask
            }
            return collated

        return collate_fn

    def __len__(self):
        return len(self.sentences or [])

    def inference(self, sentences):
        if not isinstance(sentences, list):
            sentences = [sentences]
        batch = [{'{}_seq'.format(self.source): self.processing(s)} for s in sentences]
        return self.get_collate_fn()(batch)

    def __repr__(self):
        return "TextDataset\n" + \
               json.dumps({"source": self.source,
                           "root": self.root,
                           "file": self.file,
                           "processing": self.processing,
                           "Tokenizer": {
                               "name_or_path": self.tokenizer.name_or_path,
                               "vocab_size": self.tokenizer.vocab_size,
                               "tokenizer_args": self.tokenizer_args,
                               "special_tokens": self.tokenizer.special_tokens_map_extended,
                           }}, indent=4,
                          sort_keys=False, default=str)

    def show_length(self):
        print("Plotting sequences length...")
        sentence_len, tokenizer_len = [], []
        self.tokenizer_args.update({'max_length': 512})
        for index in tqdm.tqdm(range(len(self))):
            sentence = self.__getitem__(index)['{}_seq'.format(self.source)]
            x = self.tokenizer(sentence, **self.tokenizer_args).input_ids[0]
            if self.tokenizer.sep_token_id in x:
                length = ((x == self.tokenizer.sep_token_id).nonzero(as_tuple=True)[0]).item()
            elif self.tokenizer.pad_token_id in x:
                length = ((x == self.tokenizer.pad_token_id).nonzero(as_tuple=True)[0][0]).item()
            else:
                length = len(x)
            tokenizer_len.append(length)
            sentence_len.append(len(sentence.split()))

        _, ax = plt.subplots(1, 2)
        sns.histplot(tokenizer_len, ax=ax[0], binwidth=2)
        sns.histplot(sentence_len, ax=ax[1], binwidth=2)
        ax[0].set_title("tokenizer_len")
        ax[1].set_title("sentence_len")
        plt.show()



import torch
import torch.nn as nn
from vilmedic.networks.models.utils import get_n_params

from vilmedic.networks.blocks.vision import *
from vilmedic.networks.blocks.huggingface.encoder.encoder_model import EncoderModel

from tqdm import tqdm
import numpy as np


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]

    losses = []
    linguistics = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        batch = {k: v.cuda() for k, v in batch.items()}
        out = model(**batch)
        losses.append(out['loss'].mean().cpu().data.numpy())

        if not from_training:
            linguistics.append(out['linguistic'].cpu().data)
            visuals.append(out['visual'].cpu().data)

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'linguistic': torch.cat(linguistics),
            'visual': torch.cat(visuals)
            }


class ConVIRTLoss(nn.Module):
    def __init__(self, tau, lambda_, **kwargs):
        super().__init__()
        self.tau = tau
        self.lambda_ = lambda_
        self.cos_loss = nn.CosineSimilarity()

    def forward(self, linguistic, visual):
        nominator = torch.exp(torch.div(self.cos_loss(linguistic, visual), self.tau))

        denominator_l = self.pairwise_cosine_distance(linguistic, visual)
        denominator_l = torch.sum(torch.exp(torch.div(denominator_l, self.tau)), dim=1)
        loss_l = -torch.log(torch.div(nominator, denominator_l))

        denominator_v = self.pairwise_cosine_distance(visual, linguistic)
        denominator_v = torch.sum(torch.exp(torch.div(denominator_v, self.tau)), dim=1)
        loss_v = -torch.log(torch.div(nominator, denominator_v))
        loss = torch.mean(self.lambda_ * loss_v + (1 - self.lambda_) * loss_l)
        return loss, loss_l, loss_v

    @staticmethod
    def pairwise_cosine_distance(a, b):
        eps = 1e-08
        a_n, b_n = a.norm(dim=1)[:, None], b.norm(dim=1)[:, None]
        a_norm = a / torch.clamp(a_n, min=eps)
        b_norm = b / torch.clamp(b_n, min=eps)
        return torch.mm(a_norm, b_norm.transpose(0, 1))

    def __repr__(self):
        return "ConVIRTLoss(\n" + \
               "\t(cos_loss): CosineSimilarity()\n" + \
               "\t(tau): {}\n".format(self.tau) + \
               "\t(lambda_): {}\n".format(self.lambda_) + \
               ")"


class InfoNCELoss(nn.Module):
    def __init__(self, tau, **kwargs):
        super().__init__()
        self.tau = tau
        self.xe = nn.CrossEntropyLoss(reduction='none')

    def forward(self, linguistic, visual):
        n = linguistic.shape[0]
        logits = linguistic @ visual.T
        labels = torch.arange(n).cuda()
        loss_t = self.xe(logits, labels)
        loss_i = self.xe(logits.T, labels)
        loss = (loss_i + loss_t) / 2
        loss = loss.mean()
        return loss, loss_t, loss_i

    def __repr__(self):
        return "InfoNCELoss(\n" + \
               "\t(tau): {}\n".format(self.tau) + \
               ")"


class ConVIRT(nn.Module):

    def __init__(self, encoder, cnn, projection, loss, forward_batch_size=256, **kwargs):
        super().__init__()

        # Linguistic encoder
        self.linguistic = EncoderModel(encoder)

        # Visual Encoder
        self.visual = eval(cnn.pop('proto'))(**cnn)

        # Projection
        self.vis_proj = nn.Sequential(
            nn.Linear(projection.visual_embedding_dim, projection.projection_dim),
            nn.ReLU(),
            nn.Linear(projection.projection_dim, projection.projection_dim),
        )
        self.lin_proj = nn.Sequential(
            nn.Linear(projection.textual_embedding_dim, projection.projection_dim),
            nn.ReLU(),
            nn.Linear(projection.projection_dim, projection.projection_dim),
        )

        self.loss_fn = eval(loss.pop("proto"))(**loss)
        self.fbs = forward_batch_size

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, **kwargs):
        images = images.cuda()
        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()

        bs = images.shape[0]
        linguistics = []
        visuals = []
        for i in range(int(bs / min(self.fbs, bs))):
            inp = input_ids[i * self.fbs:(i + 1) * self.fbs]
            att = attention_mask[i * self.fbs:(i + 1) * self.fbs]
            im = images[i * self.fbs:(i + 1) * self.fbs]

            linguistic = self.linguistic(input_ids=inp,
                                         attention_mask=att,
                                         **kwargs)
            linguistic = self.lin_proj(linguistic['pooler_output'])
            visual = self.vis_proj(self.visual(im))

            linguistics.append(linguistic)
            visuals.append(visual)

        linguistics = torch.cat(linguistics)
        visuals = torch.cat(visuals)

        loss, loss_l, loss_v = self.loss_fn(linguistics, visuals)

        return {"loss": loss, "loss_l": loss_l, "loss_v": loss_v, "linguistic": linguistics, "visual": visuals}

    def __repr__(self):
        s = "ConVIRT\n"
        s += str(self.visual) + '\n'
        s += str(self.linguistic) + '\n'
        s += str(self.loss_fn) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import os
import glob
import torch
import transformers
import torch.nn as nn
from omegaconf import OmegaConf
from torch.utils.data import DataLoader
from vilmedic.constants import MODEL_ZOO_CACHE_DIR
from .utils import download_model

from ..networks import *
from ..datasets import *

transformers.logging.set_verbosity_error()

MODEL_ZOO = {
    'selfsup/gloria-chexpert': ["1ibtEQH8XXNPy1Y0fE1ooSF7Jh7bdG58C", "1.6 GB"],
    'selfsup/gloria-mimic-48': ["1V50_NUWY-k2ebzmgQxRUcMOIH0UhSAEr", "1.6 GB"],
    'selfsup/convirt-mimic-balanced': ["1bTf16uRygWwTor3X0rYBhD7MuHSUjWxM", "1.4 GB"],
    'selfsup/convirt-mimic': ["1py2k6kFO1tbSlHroAKmHjOnayIakOtMm", "1.4 GB"],
    'selfsup/convirt-padchest-16': ["15p6ZaoqcxAB0dct7P9zgOcTfvZw-XkVv", "1.4 GB"],
    'selfsup/convirt-padchest-32': ["1sd9eNVDcZTPrpmlHSqjMP6hRBHyM7e4p", "1.4 GB"],
    'selfsup/convirt-indiana-16': ["10k9RlLJVLH1tuuSjzwHQK1BTmTVOUinH", "1.4 GB"],
    'selfsup/convirt-indiana-32': ["17q0MllKOnGQY7zudhK03sqDsrnf_THu2", "1.4 GB"],
    'selfsup/convirt-indiana-64': ["17xi8Mj3Ts9qeFT0l83a5Hm82W1ocoYCW", "1.4 GB"],
    'selfsup/simclr-mimic-32': ["1ibtEQH8XXNPy1Y0fE1ooSF7Jh7bdG58C", "300 MB"],
    'selfsup/simclr-mimic-64': ["1RYhQkaR9F0LbozVs7hHv0c52Js1LDh6J", "300 MB"],
    'selfsup/simclr-mimic-128': ["1w1XYaprrJrjIk-JlKpbw7OSe3sABKDkN", "300 MB"],
    'rrg/biomed-roberta-baseline-mimic': ["1aXxHkzbLdYQpLYvlQLw7NENE7LXgkc1y", "1.8 GB"],
    'rrg/biomed-roberta-baseline-indiana': ["1BzTPf4AMLF_2KGs6RX3W30HyekeUElmW", "1.8 GB"],
    'rrg/baseline-padchest': ["1COYPFZJTiG5TBlhGSX7GyswXwKL6HAW0", "1.8 GB"],
    'rrs/biomed-roberta-baseline-mimic': ["1hmEvUjKOlNsY-xipEgUZOCQm4k9mHgWR", "3.3 GB"],
    'rrs/biomed-roberta-baseline-indiana': ["1xG80gsckbdNvAVhqGo-4Lsvkwk7wy_-v", "3.3 GB"],
    'mvqa/mvqa-imageclef': ["1VmiJEGs-jYNGlbVXGi6uGmdhc06Ps4GF", "970 MB"],
}


class AutoModel:
    def __init__(self):
        raise EnvironmentError(
            "AutoModel is designed to be instantiated "
            "using the `AutoModel.from_pretrained(pretrained_model_name_or_path)` or "
            "`AutoModel.from_config(config)` methods."
        )

    @staticmethod
    def from_config(config):
        return None

    @staticmethod
    def from_pretrained(pretrained_model_name):
        try:
            file_id, _ = MODEL_ZOO[pretrained_model_name]
        except KeyError:
            raise KeyError("Unrecognized pretrained_model_name {}. "
                           "Model name should be one of {}.".format(pretrained_model_name, MODEL_ZOO.keys()))

        checkpoint_dir = os.path.join(MODEL_ZOO_CACHE_DIR, pretrained_model_name)

        if not os.path.exists(checkpoint_dir):
            print("Downloading in {}".format(MODEL_ZOO_CACHE_DIR))
            download_model(file_id=file_id, unzip_dir=checkpoint_dir)

        checkpoint = glob.glob(os.path.join(checkpoint_dir, '*.pth'))
        assert len(checkpoint) == 1, "More than one or no checkpoint found"
        state_dict = torch.load(checkpoint[0])

        try:
            config = OmegaConf.load(os.path.join(checkpoint_dir, 'config.yml'))
        except FileNotFoundError:
            raise FileNotFoundError("The file config.yml is missing")

        try:
            model_config = config["model"]
            dataset_config = config["dataset"]
        except KeyError:
            raise KeyError(
                "This config doesnt have a model and/or dataset key. Deprecated checkpoint of vilmedic version?")

        try:
            classname = dataset_config.pop("proto")
            dataset = eval(classname)(split='test', ckpt_dir=None, **dataset_config)
        except NameError:
            raise NameError(
                "Dataset {} does not exist anymore. Deprecated checkpoint of vilmedic version?".format(classname))

        try:
            classname = model_config.pop("proto")
            model: nn.Module = eval(classname)(**model_config, dl=DataLoader(dataset), logger=None)
        except NameError:
            raise NameError(
                "Model {} does not exists anymore. Deprecated checkpoint of vilmedic version?".format(classname))

        try:
            model.load_state_dict(state_dict["model"], strict=True)
        except Exception as e:
            print(e)
            raise

        model.eval()
        model.cuda()

        assert hasattr(dataset, "inference"), "Dataset has not implemented an inference function"

        print("Everything has been loaded successfully")

        return model, dataset



import torch.nn as nn
from ..rnn.utils import get_n_params, set_embeddings
import copy

# v4.3.2
from transformers.models.bert_generation import BertGenerationEncoder, BertGenerationConfig, BertGenerationDecoder
from transformers import EncoderDecoderModel, EncoderDecoderConfig

from .CNNTransformer import CNNTransformer
from .beam import beam_search


class Im2SeqHug(nn.Module):
    def __init__(self, encoder, decoder, **kwargs):
        super().__init__()

        # Encoder
        self.enc = CNNTransformer(**encoder)

        # Decoder
        decoder = vars(decoder)
        dec_config = copy.deepcopy(enc_config)
        dec_config.update(decoder)
        dec_config.is_decoder = True
        dec_config.add_cross_attention = True
        self.dec = BertGenerationDecoder(dec_config)

        # Encdec
        self.enc_dec = EncoderDecoderModel(encoder=self.enc, decoder=self.dec)
        self.bos_token_id = enc_config.bos_token_id
        self.eos_token_id = enc_config.eos_token_id
        self.pad_token_id = enc_config.pad_token_id

        # Evaluation
        self.eval_func = beam_search

        # Embeddings
        if 'src_emb' in kwargs:
            set_embeddings(kwargs['src_emb'], self.enc.embeddings.word_embeddings)
        if 'tgt_emb' in kwargs:
            set_embeddings(kwargs['tgt_emb'], self.dec.bert.embeddings.word_embeddings)

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask):
        input_ids = input_ids.cuda()
        decoder_input_ids = decoder_input_ids.cuda()
        attention_mask = attention_mask.cuda()
        decoder_attention_mask = decoder_attention_mask.cuda()

        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           labels=decoder_input_ids)
        out = vars(out)
        return out

    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s

    def get_cls(self, encoder_proto, decoder_proto):
        ret = []
        for proto in [encoder_proto, decoder_proto]:
            if 'BertGeneration' in proto:
                ret.extend([BertGenerationConfig])
                if 'Encoder' in proto:
                    ret.extend([BertGenerationEncoder])
                elif 'Decoder' in proto:
                    ret.extend([BertGenerationDecoder])
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError
        assert len(ret) == 4
        return ret



from torch.utils.data import Dataset
from torchvision.transforms import *
import os
from .utils import Labels, load_file
import json


def make_labels(root, split, file):
    return load_file(os.path.join(root, split + '.' + file))


class LabelDataset(Dataset):

    def __init__(self,
                 root=None,
                 split=None,
                 file=None,
                 ckpt_dir=None,
                 label_file=None,
                 **kwargs):

        assert split is not None, "Argument split cant be None"
        assert not (file is None and label_file is None), "Please specify a file or a label_file"

        self.root = root
        self.split = split
        self.label_file = label_file
        self.labels_map = None
        self.labels = None

        # Create a label file from data file
        if file is not None:
            self.labels = make_labels(root, split, file)
            self.label_file = os.path.join(ckpt_dir, 'labels.tok')
            if split == 'train' and not os.path.exists(self.label_file):
                Labels(self.labels).dump(self.label_file)

        # Load data file
        try:
            self.labels_map = Labels().load(self.label_file)
        except FileNotFoundError:
            raise FileNotFoundError("label file does not exists, verify path or start a training")

        # Compute labels according to data file and label file
        if file is not None:
            labels = []
            for label in self.labels:
                processed_label = self.get_processed_label(label)
                labels.append(processed_label)
            self.labels = labels

    def __len__(self):
        return len(self.labels or [])

    def __getitem__(self, index):
        return {'label': self.labels[index]}

    def inference(self, label):
        if not isinstance(label, list):
            label = [label]
        batch = [{'label': self.get_processed_label(l)} for l in label]
        return self.get_collate_fn()(batch)

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {'labels': torch.stack([s['label'] for s in batch])}
            return collated

        return collate_fn

    def get_processed_label(self, label):
        try:
            classes = [l for l in label.split(',')]
            if not self.labels_map.multi_label:  # single label
                return torch.tensor(self.labels_map.label2idx[classes[0]]).long()
            else:  # multi-label
                multi_hot = torch.zeros(len(self.labels_map.idx2label))
                multi_hot[[self.labels_map.label2idx[c] for c in classes]] = 1.
                return multi_hot
        except KeyError:
            # Label in a split is not present in train set, this can happen (OOD or so.)
            return torch.tensor(-100).long()

    def __repr__(self):
        return "LabelDataset\n" + \
               json.dumps({"num_labels": len(self.labels_map.labels)}, indent=4,
                          sort_keys=False, default=str)



from transformers import BertConfig
from transformers.models.bert.modeling_bert import BertPreTrainedModel, BertEncoder, BertPooler, BertEmbeddings, \
    BertPredictionHeadTransform
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
import torch.nn as nn
import torch
from copy import deepcopy
import torch.nn.functional as F

BertLayerNorm = torch.nn.LayerNorm


class LogitBinaryCrossEntropy(nn.Module):
    """Returns Binary Cross Entropy for logits.
    Attention:
        `Key`: logit_bce
    """

    def __init__(self):
        super().__init__()

    def forward(self, sample_list, model_output):
        """Calculates and returns the binary cross entropy for logits
        Args:
            sample_list (SampleList): SampleList containing `targets` attribute.
            model_output (Dict): Model output containing `scores` attribute.
        Returns:
            torch.FloatTensor: Float value for loss.
        """
        scores = model_output["scores"]
        targets = sample_list["targets"]
        loss = F.binary_cross_entropy_with_logits(scores, targets, reduction="mean")

        return loss * targets.size(1)


class BertVisioLinguisticEmbeddings(BertEmbeddings):
    def __init__(self, config, *args, **kwargs):
        super().__init__(config)
        self.token_type_embeddings_visual = nn.Embedding(
            config.type_vocab_size, config.hidden_size
        )
        self.position_embeddings_visual = nn.Embedding(
            config.max_position_embeddings, config.hidden_size
        )

        self.projection = nn.Linear(config.visual_embedding_dim, config.hidden_size)
        # self.word_embeddings = nn.Embedding(config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id)

    def initialize_visual_from_pretrained(self):
        self.token_type_embeddings_visual.weight = nn.Parameter(
            deepcopy(self.token_type_embeddings.weight.data), requires_grad=True
        )
        self.position_embeddings_visual.weight = nn.Parameter(
            deepcopy(self.position_embeddings.weight.data), requires_grad=True
        )

    def encode_text(
            self, input_ids, token_type_ids=None
    ):
        seq_length = input_ids.size(1)
        position_ids = torch.arange(
            seq_length, dtype=torch.long, device=input_ids.device
        )
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)
        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        words_embeddings = self.word_embeddings(input_ids)

        position_embeddings = self.position_embeddings(position_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)
        embeddings = words_embeddings + position_embeddings + token_type_embeddings
        return embeddings

    def encode_image(
            self,
            visual_embeddings,
            visual_embeddings_type,
            image_text_alignment=None,
    ):

        visual_embeddings = self.projection(visual_embeddings)
        token_type_embeddings_visual = self.token_type_embeddings_visual(
            visual_embeddings_type
        )

        # get position_embeddings
        # this depends on image_text_alignment
        position_embeddings_visual = self.get_position_embeddings_visual(
            visual_embeddings, image_text_alignment=image_text_alignment
        )

        # calculate visual embeddings
        v_embeddings = (
                visual_embeddings
                + position_embeddings_visual
                + token_type_embeddings_visual
        )
        return v_embeddings

    def get_position_embeddings_visual(
            self, visual_embeddings, image_text_alignment=None
    ):

        if image_text_alignment is not None:
            # image_text_alignment = Batch x image_length x alignment_number.
            # Each element denotes the position of the word corresponding to the
            # image feature. -1 is the padding value.
            image_text_alignment_mask = (
                (image_text_alignment != -1).long().to(image_text_alignment.device)
            )
            # Get rid of the -1.
            image_text_alignment = image_text_alignment_mask * image_text_alignment

            # position_embeddings_visual
            # = Batch x image_length x alignment length x dim
            position_embeddings_visual = self.position_embeddings(
                image_text_alignment
            ) * image_text_alignment_mask.unsqueeze(-1)
            position_embeddings_visual = position_embeddings_visual.sum(2)

            # We want to averge along the alignment_number dimension.
            image_text_alignment_mask = image_text_alignment_mask.sum(2)
            image_text_alignment_mask[image_text_alignment_mask == 0] = torch.tensor(
                [1], dtype=torch.long
            )  # Avoid devide by zero error
            position_embeddings_visual = (
                    position_embeddings_visual / image_text_alignment_mask.unsqueeze(-1)
            )

            position_ids_visual = torch.zeros(
                visual_embeddings.size()[:-1],
                dtype=torch.long,
                device=visual_embeddings.device,
            )

            position_embeddings_visual = (
                    position_embeddings_visual
                    + self.position_embeddings_visual(position_ids_visual)
            )
        else:
            position_ids_visual = torch.zeros(
                visual_embeddings.size()[:-1],
                dtype=torch.long,
                device=visual_embeddings.device,
            )
            position_embeddings_visual = self.position_embeddings_visual(
                position_ids_visual
            )

        return position_embeddings_visual

    def forward(
            self,
            input_ids,
            token_type_ids=None,
            visual_embeddings=None,
            visual_embeddings_type=None,
            image_text_alignment=None,
            **kwargs
    ):
        """
        input_ids = [batch_size, sequence_length]
        token_type_ids = [batch_size, sequence_length]
        visual_embedding = [batch_size, image_feature_length, image_feature_dim]
        image_text_alignment = [batch_size, image_feature_length, alignment_dim]
        """

        # text embeddings
        text_embeddings = self.encode_text(input_ids, token_type_ids=token_type_ids)

        # visual embeddings
        if visual_embeddings is not None and visual_embeddings_type is not None:
            v_embeddings = self.encode_image(
                visual_embeddings,
                visual_embeddings_type=visual_embeddings_type,
                image_text_alignment=image_text_alignment,
            )

            # Concate the two:
            embeddings = torch.cat(
                (text_embeddings, v_embeddings), dim=1
            )  # concat the visual embeddings after the attentions

        else:
            embeddings = text_embeddings

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class VisualBERTBase(nn.Module):
    def __init__(
            self,
            num_labels=None,
            visual_embedding_dim=None,
            **kwargs
    ):
        super().__init__()
        assert visual_embedding_dim is not None
        assert num_labels is not None
        config = BertConfig.from_pretrained('bert-base-uncased',
                                            num_labels=num_labels,
                                            **kwargs)


        self.config = config
        self.config.visual_embedding_dim = visual_embedding_dim

        self.embeddings = BertVisioLinguisticEmbeddings(self.config)

        # self.embeddings = nn.Embedding(config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id)

        # to remove
        # self.lstm = nn.LSTM(
        #     input_size=self.config.hidden_size,
        #     hidden_size=self.config.hidden_size,
        #     num_layers=1,
        #     batch_first=True
        # )

        self.proj = nn.Linear(self.config.visual_embedding_dim, config.hidden_size)

        self.encoder = BertEncoder(self.config)
        self.pooler = BertPooler(self.config)

        self.dropout = nn.Dropout(self.config.hidden_dropout_prob)
        self.classifier = nn.Sequential(
            BertPredictionHeadTransform(self.config),
            nn.Linear(self.config.hidden_size, num_labels)
        )

        self.init_weights()

        self.loss_fct = nn.CrossEntropyLoss()
        self.num_labels = num_labels

    def init_weights(self):
        self.apply(self._init_weights)

    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
            self,
            input_ids,
            labels,
            input_mask=None,
            attention_mask=None,
            visual_embeddings=None,
            token_type_ids=None,
            visual_embeddings_type=None,
            image_text_alignment=None,
    ):

        image_mask = torch.ones(
            visual_embeddings.size(-2), device=visual_embeddings.device
        ).expand(visual_embeddings.size()[:-1]).long()

        attention_mask = torch.cat((input_mask, image_mask), dim=-1)

        if visual_embeddings_type is None:
            visual_embeddings_type = torch.zeros_like(image_mask)

        extended_attention_mask = attention_mask.unsqueeze(1).unsqueeze(2)

        if not torch.jit.is_scripting():
            extended_attention_mask = extended_attention_mask.to(
                dtype=next(self.parameters()).dtype
            )  # fp16 compatibility

        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        embeddings = self.embeddings(
            input_ids,
            token_type_ids,
            visual_embeddings=visual_embeddings,
            visual_embeddings_type=visual_embeddings_type,
            image_text_alignment=image_text_alignment,
        )

        # text_embeddings = self.embeddings(input_ids)
        # text_embeddings, _ = self.lstm(text_embeddings)
        # v_embeddings = self.proj(visual_embeddings)
        # embeddings = torch.cat(
        #     (text_embeddings, v_embeddings), dim=1
        # )

        # Only keep last layer hidden states (no output attentions)
        encoded_layers = self.encoder(embeddings, extended_attention_mask)
        sequence_output = encoded_layers[0]

        # pooled_output = self.pooler(sequence_output)
        attn_data_list = encoded_layers[1:]

        # # VQA
        index_to_gather = input_mask.sum(1) - 2
        pooled_output = torch.gather(
            sequence_output,
            1,
            index_to_gather.unsqueeze(-1).unsqueeze(-1).expand(index_to_gather.size(0), 1, sequence_output.size(-1)),
        )

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        reshaped_logits = logits.contiguous().view(-1, self.num_labels)

        output_dict = {"logits": reshaped_logits,
                       "loss": self.loss_fct(reshaped_logits, labels)}
        return output_dict



# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

from .layer_norm import LayerNorm
from .fc import FC, MLP

import torch.nn as nn
import torch.nn.functional as F
import torch
import math


# ------------------------------
# ---- Multi-Head Attention ----
# ------------------------------

class MHAtt(nn.Module):
    def __init__(self, __C):
        super(MHAtt, self).__init__()
        self.__C = __C

        self.linear_v = nn.Linear(__C.HIDDEN_SIZE, __C.HIDDEN_SIZE)
        self.linear_k = nn.Linear(__C.HIDDEN_SIZE, __C.HIDDEN_SIZE)
        self.linear_q = nn.Linear(__C.HIDDEN_SIZE, __C.HIDDEN_SIZE)
        self.linear_merge = nn.Linear(__C.HIDDEN_SIZE, __C.HIDDEN_SIZE)

        self.dropout = nn.Dropout(__C.DROPOUT_R)

    def forward(self, v, k, q, mask):
        n_batches = q.size(0)

        v = self.linear_v(v).view(
            n_batches,
            -1,
            self.__C.MULTI_HEAD,
            int(self.__C.HIDDEN_SIZE / self.__C.MULTI_HEAD)
        ).transpose(1, 2)

        k = self.linear_k(k).view(
            n_batches,
            -1,
            self.__C.MULTI_HEAD,
            int(self.__C.HIDDEN_SIZE / self.__C.MULTI_HEAD)
        ).transpose(1, 2)

        q = self.linear_q(q).view(
            n_batches,
            -1,
            self.__C.MULTI_HEAD,
            int(self.__C.HIDDEN_SIZE / self.__C.MULTI_HEAD)
        ).transpose(1, 2)

        atted = self.att(v, k, q, mask)
        atted = atted.transpose(1, 2).contiguous().view(
            n_batches,
            -1,
            self.__C.HIDDEN_SIZE
        )

        atted = self.linear_merge(atted)

        return atted

    def att(self, value, key, query, mask):
        d_k = query.size(-1)

        scores = torch.matmul(
            query, key.transpose(-2, -1)
        ) / math.sqrt(d_k)

        if mask is not None:
            scores = scores.masked_fill(mask, -1e9)

        att_map = F.softmax(scores, dim=-1)
        att_map = self.dropout(att_map)

        return torch.matmul(att_map, value)


# ---------------------------
# ---- Feed Forward Nets ----
# ---------------------------

class FFN(nn.Module):
    def __init__(self, __C):
        super(FFN, self).__init__()

        self.mlp = MLP(
            in_size=__C.HIDDEN_SIZE,
            mid_size=__C.FF_SIZE,
            out_size=__C.HIDDEN_SIZE,
            dropout_r=__C.DROPOUT_R,
            use_relu=True
        )

    def forward(self, x):
        return self.mlp(x)


# ------------------------
# ---- Self Attention ----
# ------------------------

class SA(nn.Module):
    def __init__(self, __C):
        super(SA, self).__init__()

        self.mhatt = MHAtt(__C)
        self.ffn = FFN(__C)

        self.dropout1 = nn.Dropout(__C.DROPOUT_R)
        self.norm1 = LayerNorm(__C.HIDDEN_SIZE)

        self.dropout2 = nn.Dropout(__C.DROPOUT_R)
        self.norm2 = LayerNorm(__C.HIDDEN_SIZE)

    def forward(self, y, y_mask):
        y = self.norm1(y + self.dropout1(
            self.mhatt(y, y, y, y_mask)
        ))

        y = self.norm2(y + self.dropout2(
            self.ffn(y)
        ))

        return y


# -------------------------------
# ---- Self Guided Attention ----
# -------------------------------

class SGA(nn.Module):
    def __init__(self, __C):
        super(SGA, self).__init__()

        self.mhatt1 = MHAtt(__C)
        self.mhatt2 = MHAtt(__C)
        self.ffn = FFN(__C)

        self.dropout1 = nn.Dropout(__C.DROPOUT_R)
        self.norm1 = LayerNorm(__C.HIDDEN_SIZE)

        self.dropout2 = nn.Dropout(__C.DROPOUT_R)
        self.norm2 = LayerNorm(__C.HIDDEN_SIZE)

        self.dropout3 = nn.Dropout(__C.DROPOUT_R)
        self.norm3 = LayerNorm(__C.HIDDEN_SIZE)

    def forward(self, x, y, x_mask, y_mask):
        x = self.norm1(x + self.dropout1(
            self.mhatt1(v=x, k=x, q=x, mask=x_mask)
        ))

        x = self.norm2(x + self.dropout2(
            self.mhatt2(v=y, k=y, q=x, mask=y_mask)
        ))

        x = self.norm3(x + self.dropout3(
            self.ffn(x)
        ))

        return x


# ------------------------------------------------
# ---- MAC Layers Cascaded by Encoder-Decoder ----
# ------------------------------------------------

class MCA_ED(nn.Module):
    def __init__(self, __C):
        super(MCA_ED, self).__init__()

        self.enc_list = nn.ModuleList([SA(__C) for _ in range(__C.LAYER)])
        self.dec_list = nn.ModuleList([SGA(__C) for _ in range(__C.LAYER)])

    def forward(self, y, x, y_mask, x_mask):
        # Get encoder last hidden vector
        for enc in self.enc_list:
            y = enc(y, y_mask)

        # Input encoder last hidden vector
        # And obtain decoder last hidden vectors
        for dec in self.dec_list:
            x = dec(x, y, x_mask, y_mask)

        return y, x



import torch.nn as nn
import copy

# v4.3.2
from transformers.models.bert_generation import BertGenerationEncoder, BertGenerationConfig, BertGenerationDecoder
from transformers import EncoderDecoderModel as HFEncoderDecoderModel
from vilmedic.networks.models.utils import get_n_params


class EncoderDecoderModel(nn.Module):
    """
    If proto is mentioned in encoder and decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder/BertGenerationDecoder model from encoder and decoder dict.
    """

    def __init__(self, encoder, decoder, **kwargs):
        super().__init__()
        if 'proto' in decoder and 'proto' in encoder:
            self.enc_dec = HFEncoderDecoderModel.from_encoder_decoder_pretrained(encoder.pop('proto'),
                                                                                 decoder.pop('proto'))
        else:
            # Encoder
            enc_config = BertGenerationConfig(**encoder)
            enc = BertGenerationEncoder(enc_config)

            # Decoder
            dec_config = copy.deepcopy(enc_config)
            dec_config.update(decoder)
            dec_config.is_decoder = True
            dec_config.add_cross_attention = True
            dec = BertGenerationDecoder(dec_config)

            # Encdec
            self.enc_dec = HFEncoderDecoderModel(encoder=enc, decoder=dec)

        assert self.enc_dec.config.is_encoder_decoder == True

        # Inference
        self.generate = self.enc_dec.generate
        self.config = self.enc_dec.config

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask):
        input_ids = input_ids.cuda()
        decoder_input_ids = decoder_input_ids.cuda()
        attention_mask = attention_mask.cuda()
        decoder_attention_mask = decoder_attention_mask.cuda()
        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           labels=decoder_input_ids)
        out = vars(out)
        return out

    def __repr__(self):
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
from torch.utils.data import Dataset

from .base.LabelDataset import LabelDataset
from .ImSeq import ImSeq


class ImSeqLabel(Dataset):
    def __init__(self, seq, label, image, split, ckpt_dir, **kwargs):
        self.split = split
        self.imgseq = ImSeq(seq, image, split=split, ckpt_dir=ckpt_dir)
        self.label = LabelDataset(**label, split=split, ckpt_dir=ckpt_dir)

        assert len(self.imgseq) == len(self.label), str(len(self.imgseq)) + 'vs ' + str(len(self.label))

        # For decoding, if needed
        self.tokenizer = self.imgseq.seq.tokenizer
        self.tokenizer_max_len = self.imgseq.seq.tokenizer_max_len

        # For tokenizing
        self.tokenizer_args = self.imgseq.seq.tokenizer_args

    def __getitem__(self, index):
        return {**self.imgseq.__getitem__(index), **self.label.__getitem__(index)}

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {**self.imgseq.get_collate_fn()(batch),
                        **self.label.get_collate_fn()(batch)}
            return collated

        return collate_fn

    def __len__(self):
        return len(self.label)

    def __repr__(self):
        return "ImSeqLabel\n" + str(self.imgseq) + '\n' + str(self.label)



import torch
import torch.nn as nn
from vilmedic.networks.models.utils import get_n_params
from pytorch_metric_learning.losses import NTXentLoss

from vilmedic.networks.blocks.vision import *

from tqdm import tqdm
import numpy as np


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]
    losses = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for batch in pbar:
        out = model(**batch, from_training=from_training)
        losses.append(out['loss'].mean().cpu().data.numpy())
        if not from_training:
            visuals.append(out['visual'].cpu().data)

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'visual': torch.cat(visuals)
            }


class SimCLR(nn.Module):

    def __init__(self, cnn, projection, loss, forward_batch_size=256, **kwargs):
        super().__init__()

        # Visual Encoder
        self.visual = eval(cnn.pop('proto'))(**cnn)

        # Projection
        self.vis_proj = nn.Sequential(
            nn.Linear(projection.visual_embedding_dim, projection.projection_dim),
            nn.ReLU(),
            nn.Linear(projection.projection_dim, projection.projection_dim),
        )

        self.loss_fn = NTXentLoss(temperature=loss.tau)
        self.fbs = forward_batch_size

        # Evaluation
        self.eval_func = evaluation

    def forward(self, images, from_training=True, **kwargs):
        if from_training:
            images = torch.cat(torch.split(images, [3, 3], dim=1), dim=0)

        # forward passes
        visuals = []
        images = images.cuda()
        bs = images.shape[0]
        for i in range(int(bs / min(self.fbs, bs))):
            im = images[i * self.fbs:(i + 1) * self.fbs]
            visual = self.vis_proj(self.visual(im))
            visuals.append(visual)
        visuals = torch.cat(visuals)

        # if from ensemblor, return embeddings
        if not from_training:
            return {"loss": torch.tensor(0.), "visual": visuals}

        # Computing loss
        batch_size = visuals.shape[0] // 2
        indices = torch.arange(0, batch_size, device=visuals.device)
        labels = torch.cat((indices, indices))
        loss = self.loss_fn(visuals, labels)

        return {"loss": loss}

    def __repr__(self):
        s = "SimCLR\n"
        s += str(self.visual) + '\n'
        s += str(self.loss_fn) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import os
import itertools


def load_file(path):
    """Default loading function, which loads nth sentence at line n.
    """
    with open(path, 'r') as f:
        content = f.read().strip()
    return [s for s in content.split('\n')]


class Vocab:
    def __init__(self, sentences=None,
                 pad_token="[PAD]",
                 eos_token="[SEP]",
                 bos_token="[CLS]",
                 unk_token="[UNK]",
                 mask_token="[MASK]"):
        self.vocab = list(itertools.chain(*sentences))
        self.words = [bos_token, pad_token, eos_token, unk_token, mask_token] + sorted(set(self.vocab))

    def dump(self, path):
        open(path, "w").write("\n".join(str(w) for w in self.words))


class Labels:
    def __init__(self, labels=None):
        if labels is not None:
            self.labels = list(set([l for label in labels for l in label.split(',')]))
            self.multi_label = max([len(label.split(',')) for label in labels]) > 1

    def dump(self, path):
        open(path, "w").write("\n".join(str(w) for w in ["multi-label:" + str(self.multi_label)] + self.labels))

    def load(self, path):
        self.labels = [w.strip() for w in open(path, "r").readlines()]
        self.multi_label = eval(self.labels.pop(0).split(':')[-1])
        assert isinstance(self.multi_label, bool), 'Bad formatting'
        self.label2idx = {l: i for i, l in enumerate(self.labels)}
        self.idx2label = {i: l for i, l in enumerate(self.labels)}
        return self



# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

import torch.nn as nn
import torch


class FC(nn.Module):
    def __init__(self, in_size, out_size, dropout_r=0., use_relu=True):
        super(FC, self).__init__()
        self.dropout_r = dropout_r
        self.use_relu = use_relu

        self.linear = nn.Linear(in_size, out_size)

        if use_relu:
            self.relu = nn.ReLU(inplace=True)

        if dropout_r > 0:
            self.dropout = nn.Dropout(dropout_r)

    def forward(self, x):
        x = self.linear(x)

        if self.use_relu:
            x = self.relu(x)

        if self.dropout_r > 0:
            x = self.dropout(x)

        return x


class MLP(nn.Module):
    def __init__(self, in_size, mid_size, out_size, dropout_r=0., use_relu=True):
        super(MLP, self).__init__()

        self.fc = FC(in_size, mid_size, dropout_r=dropout_r, use_relu=use_relu)
        self.linear = nn.Linear(mid_size, out_size)

    def forward(self, x):
        return self.linear(self.fc(x))



import collections
import argparse
import logging
import copy
import json
import time
import random
import torch
import os
import numpy as np
from omegaconf import OmegaConf
from random import randrange
import re


def extract_seed_from_ckpt(ckpt):
    # 1.68_10_560435.pth to 560435
    assert os.path.exists(ckpt), '{} does not exist'.format(ckpt)
    return re.match(".*_(.*?).pth", ckpt).group(1)


def print_args(config, splits, seed, override):
    logger = logging.getLogger(str(seed))
    logger.settings("Override dict")
    logger.info(json.dumps(OmegaConf.to_container(override), indent=4, sort_keys=True))

    for split in splits:
        d = OmegaConf.to_container(getattr(config, split))
        logger.settings(split)
        logger.info(json.dumps(d, indent=4, sort_keys=True))


def get_args():
    """Return Omegaconf dict from command line
    Command line consists of one config and ovveriding args (others)"""

    parser = argparse.ArgumentParser()
    parser.add_argument('config', type=str)
    args, others = parser.parse_known_args()

    # Get configs
    config = OmegaConf.load(args.config)
    includes = config.get("includes", [])

    if not isinstance(includes, collections.abc.Sequence):
        raise AttributeError(
            "Includes must be a list, {} provided".format(type(includes))
        )

    # Loop over includes
    include_mapping = OmegaConf.create()
    for include in includes:
        if not os.path.exists(include):
            include = os.path.join(os.path.dirname(args.config), include)

        current_include_mapping = OmegaConf.load(include)
        include_mapping = OmegaConf.merge(include_mapping, current_include_mapping)

    # Override includes with current config
    config = OmegaConf.merge(include_mapping, config)

    # Override current config with additional args
    override = OmegaConf.from_dotlist(others)
    config = OmegaConf.merge(config, override)

    return config, override


def get(config, mode):
    """ Create a personal dict for each executor (trainor, validator or ensemblor)
    For an executor, we give all args but the other executors args
    """
    exec_config = copy.deepcopy(getattr(config, mode))
    for att in list(config.keys()):
        if att not in ['trainor', 'validator', 'ensemblor']:
            exec_config[att] = config[att]
    return exec_config


def get_seed(seed=None):
    if seed is None:
        seed = randrange(100000, 999999)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.benchmark = True
    return seed



import torch.nn as nn

import torch
from vilmedic.networks.blocks.huggingface.encoder_decoder.evaluation import evaluation
from vilmedic.networks.models.utils import get_n_params

# v4.3.2
from transformers.modeling_outputs import Seq2SeqLMOutput
from vilmedic.networks.blocks.huggingface.encoder_decoder.encoder_decoder_model import EncoderDecoderModel


class MultimodalEnc(nn.Module):
    def __init__(self, encoder, cnn):
        super().__init__()
        self.encoder = encoder

        cnn_func = cnn.pop('proto')
        self.visual_projection = nn.Linear(cnn.pop("visual_embedding_dim"), self.encoder.config.hidden_size)
        self.cnn = eval(cnn_func)(**cnn)

    def forward(self, input_ids, images, **kwargs):
        # Encoder
        encoder_outputs = self.encoder(input_ids, **kwargs)

        # CNN
        with torch.no_grad():
            visual_features = self.cnn(images.cuda())

        # Add visual attribute to encoder_outputs
        visual_features = self.visual_projection(visual_features)
        encoder_outputs.visual_features = visual_features

        return encoder_outputs

    def train(self, mode: bool = True):
        self.training = mode
        for module in self.children():
            module.train(mode)
        self.cnn.train(False)
        return self


class MultimodalEncDec(EncoderDecoderModel):
    def __init__(self, encoder, decoder, cnn, **kwargs):
        enc_dec = EncoderDecoderModel.from_encoder_decoder_pretrained(encoder.pop('proto'),
                                                                      decoder.pop('proto'))

        super().__init__(encoder=enc_dec.encoder, decoder=enc_dec.decoder, config=enc_dec.config)
        self.encoder = MultimodalEnc(self.encoder, cnn)

        # beam param
        self.to_tile = ["last_hidden_state", "visual_features"]

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            decoder_input_ids=None,
            decoder_attention_mask=None,
            encoder_outputs=None,
            past_key_values=None,
            inputs_embeds=None,
            decoder_inputs_embeds=None,
            labels=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
            **kwargs,
    ):
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            kwargs_encoder = {argument: value for argument, value in kwargs.items() if
                              not argument.startswith("decoder_")}

            # Encoder
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                **kwargs_encoder,
            )

        visual_features = encoder_outputs.visual_features
        # Concat modalities
        encoder_hidden_states = torch.cat((encoder_outputs.last_hidden_state, visual_features), dim=1)

        # update mask accordingly
        image_mask = torch.ones(
            visual_features.size(-2), device=visual_features.device
        ).expand(visual_features.size()[:-1]).long()
        attention_mask = torch.cat((attention_mask, image_mask), dim=-1)

        # Decode
        kwargs_decoder = {
            argument[len("decoder_"):]: value for argument, value in kwargs.items() if argument.startswith("decoder_")
        }
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=attention_mask,
            inputs_embeds=decoder_inputs_embeds,
            labels=labels,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            use_cache=use_cache,
            past_key_values=past_key_values,
            return_dict=return_dict,
            **kwargs_decoder,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqLMOutput(
            loss=decoder_outputs.loss,
            logits=decoder_outputs.logits,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )


class SumHugMulti(nn.Module):

    def __init__(self, encoder, decoder, cnn, **kwargs):
        super().__init__()

        self.cnn = eval(cnn.pop('proto'))(**cnn)
        self.visual_projection = nn.Linear(cnn.pop("visual_embedding_dim"), self.encoder.config.hidden_size)

        self.enc_dec = EncoderDecoderModel(encoder, decoder)
        self.enc = self.enc_dec.enc

        # Evaluation
        self.eval_func = evaluation
        self.enc_dec.to_tile = ["last_hidden_state", "visual_features"]

        self.bos_token_id = self.enc_dec.dec.config.bos_token_id
        self.eos_token_id = self.enc_dec.dec.config.eos_token_id
        self.pad_token_id = self.enc_dec.dec.config.pad_token_id


    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            decoder_input_ids=None,
            decoder_attention_mask=None,
            encoder_outputs=None,
            past_key_values=None,
            inputs_embeds=None,
            decoder_inputs_embeds=None,
            labels=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
            **kwargs,
    ):
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            kwargs_encoder = {argument: value for argument, value in kwargs.items() if
                              not argument.startswith("decoder_")}

            # Encoder
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                **kwargs_encoder,
            )

        visual_features = encoder_outputs.visual_features
        # Concat modalities
        encoder_hidden_states = torch.cat((encoder_outputs.last_hidden_state, visual_features), dim=1)

        # update mask accordingly
        image_mask = torch.ones(
            visual_features.size(-2), device=visual_features.device
        ).expand(visual_features.size()[:-1]).long()
        attention_mask = torch.cat((attention_mask, image_mask), dim=-1)

        # Decode
        kwargs_decoder = {
            argument[len("decoder_"):]: value for argument, value in kwargs.items() if argument.startswith("decoder_")
        }
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=attention_mask,
            inputs_embeds=decoder_inputs_embeds,
            labels=labels,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            use_cache=use_cache,
            past_key_values=past_key_values,
            return_dict=return_dict,
            **kwargs_decoder,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqLMOutput(
            loss=decoder_outputs.loss,
            logits=decoder_outputs.logits,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )

    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
from tqdm import tqdm


def beam_search(models, config, dl, lp_alpha=0.0):
    def tile_ctx_dict(ctx_dict, idxs):
        """Returns dict of 3D tensors repeatedly indexed along the sample axis."""
        # 1st: tensor, 2nd optional mask
        return {
            k: (t[:, idxs], None if mask is None else mask[:, idxs])
            for k, (t, mask) in ctx_dict.items()
        }

    def check_context_ndims(ctx_dict):
        bs = []
        for name, (ctx, mask) in ctx_dict.items():
            assert ctx.dim() == 3, \
                f"{name}'s 1st dim should always be a time dimension."
            bs.append(ctx.size()[1])
        # Making sure batch_size is correct for all samples
        assert len(set(bs)) == 1
        return set(bs).pop()

    # This is the batch-size requested by the user but with sorted
    # batches, efficient batch-size will be <= max_batch_size
    max_batch_size = config.batch_size
    k = config.beam_width
    inf = -1000
    max_len = dl.dataset.tgt_len

    results = []

    # For classical models that have single encoder, decoder and
    # target vocabulary
    decs = [m.dec for m in models]
    f_inits = [dec.f_init for dec in decs]
    f_nexts = [dec.f_next for dec in decs]

    # Common parts
    encoders = [m.encode for m in models]
    # vocab = dl.dataset.tgt_vocab
    # n_vocab = len(vocab)
    # eos = vocab.word2idx(vocab.get_eos())
    # unk = vocab.word2idx(vocab.get_unk())
    # bos = vocab.word2idx(vocab.get_bos())

    #
    tgt_tokenizer = dl.dataset.tgt_tokenizer
    n_vocab = tgt_tokenizer.vocab_size
    unk = tgt_tokenizer.vocab.get(tgt_tokenizer.unk_token)
    bos = tgt_tokenizer.vocab.get(tgt_tokenizer.cls_token)
    eos = tgt_tokenizer.vocab.get(tgt_tokenizer.sep_token)

    # Tensorized beam that will shrink and grow up to max_batch_size
    beam_storage = torch.zeros(
        max_len, max_batch_size, k, dtype=torch.long).cuda()
    mask = torch.arange(max_batch_size * k).cuda()
    nll_storage = torch.zeros(max_batch_size).cuda()

    # store refs
    refs = []
    pbar = tqdm(dl, total=len(dl))
    for batch in pbar:
        # Get refs
        refs.extend(batch['decoder_input_ids'].tolist())

        # Encode source modalities
        batch = {k: v.cuda() for k, v in batch.items()}
        ctx_dicts = [encode(**batch) for encode in encoders]

        # Sanity check one of the context dictionaries for dimensions and return batch_size
        batch_size = check_context_ndims(ctx_dicts[0])

        # Always use the initial storage
        beam = beam_storage.narrow(1, 0, batch_size).zero_()

        # Mask to apply to pdxs.view(-1) to fix indices
        nk_mask = mask.narrow(0, 0, batch_size * k)

        # nll: batch_size x 1 (will get expanded further)
        nll = nll_storage.narrow(0, 0, batch_size).unsqueeze(1)

        # Tile indices to use in the loop to expand first dim
        tile = range(batch_size)

        # Get initial decoder state (N*H)
        h_ts = [f_init(ctx_dict) for f_init, ctx_dict in zip(f_inits, ctx_dicts)]

        # we always have <bos> tokens except that the returned embeddings
        # may differ from one model to another.

        idxs = torch.LongTensor(batch_size).fill_(bos).cuda()

        for tstep in range(max_len):
            # Select correct positions from source context
            ctx_dicts = [tile_ctx_dict(cd, tile) for cd in ctx_dicts]

            # Get log probabilities and next state
            # log_p: batch_size x vocab_size (t = 0)
            #        batch_size*beam_size x vocab_size (t > 0)
            # NOTE: get_emb does not exist in some models, fix this.
            log_ps, h_ts = zip(
                *[f_next(cd, dec.emb(idxs), h_t[tile]) for
                  f_next, dec, cd, h_t in zip(f_nexts, decs, ctx_dicts, h_ts)])

            # Do the actual averaging of log-probabilities
            log_p = sum(log_ps).data

            # Detect <eos>'d hyps
            idxs = torch.nonzero(idxs == eos)
            if idxs.numel():
                if idxs.numel() == batch_size * k:
                    break
                idxs.squeeze_(-1)
                # Unfavor all candidates
                log_p.index_fill_(0, idxs, inf)
                # Favor <eos> so that it gets selected
                log_p.view(-1).index_fill_(0, idxs * n_vocab + 2, 0)

            # Expand to 3D, cross-sum scores and reduce back to 2D
            # log_p: batch_size x vocab_size ( t = 0 )
            #   nll: batch_size x beam_size (x 1)
            # nll becomes: batch_size x beam_size*vocab_size here
            # Reduce (N, K*V) to k-best
            nll, beam[tstep] = nll.unsqueeze_(2).add(log_p.view(
                batch_size, -1, n_vocab)).view(batch_size, -1).topk(
                k, sorted=False, largest=True)

            # previous indices into the beam and current token indices
            pdxs = beam[tstep] // n_vocab
            beam[tstep].remainder_(n_vocab)
            idxs = beam[tstep].view(-1)

            # Compute correct previous indices
            # Mask is needed since we're in flattened regime
            tile = pdxs.view(-1) + (nk_mask // k) * (k if tstep else 1)

            if tstep > 0:
                # Permute all hypothesis history according to new order
                beam[:tstep] = beam[:tstep].gather(2, pdxs.repeat(tstep, 1, 1))

        # Put an explicit <eos> to make idxs_to_sent happy
        beam[max_len - 1] = eos

        # Find lengths by summing tokens not in (pad,bos,eos)
        len_penalty = beam.gt(2).float().sum(0).clamp(min=1)

        if lp_alpha > 0.:
            len_penalty = ((5 + len_penalty) ** lp_alpha) // 6 ** lp_alpha

        # Apply length normalization
        nll.div_(len_penalty)

        # Get best-1 hypotheses
        top_hyps = nll.topk(1, sorted=False, largest=True)[1].squeeze(1)
        hyps = beam[:, range(batch_size), top_hyps].t().to('cpu')
        results.extend(hyps.tolist())

    # Recover order of the samples if necessary
    if getattr(dl.batch_sampler, 'store_indices', False):
        results = [results[i] for i, j in sorted(
            enumerate(dl.batch_sampler.orig_idxs), key=lambda k: k[1])]

    hyps = [tgt_tokenizer.decode(hyp, skip_special_tokens=True, clean_up_tokenization_spaces=False) for hyp in results]
    refs = [tgt_tokenizer.decode(ref, skip_special_tokens=True, clean_up_tokenization_spaces=False) for ref in refs]

    return {'refs': refs, 'hyps': hyps}



from transformers import BertConfig
from transformers.models.bert.modeling_bert import BertPreTrainedModel, BertEncoder, BertPooler
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
import torch.nn as nn
import torch

BertLayerNorm = torch.nn.LayerNorm


class VisioEmbeddings(nn.Module):
    """Construct the embeddings from word, position and token_type embeddings.
    """

    def __init__(self, config):
        super().__init__()
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, input_ids=None, **kwargs):
        # Inputs_ids are of size bs x spatial x feat_dim
        embeddings = input_ids
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class ImageFeaturesBert(BertPreTrainedModel):
    def __init__(self, **kwargs):
        config = BertConfig()
        super().__init__(config)

        self.config = config
        self.embeddings = VisioEmbeddings(config)
        config.num_hidden_layers=1
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)

        self.init_weights()

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            past_key_values=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
    ):
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        use_cache = False

        input_shape = input_ids.size()
        batch_size, seq_length, _ = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        # past_key_values_length
        past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0
        attention_mask = torch.ones(((batch_size, seq_length + past_key_values_length)), device=device)

        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape, device)
        encoder_extended_attention_mask = None
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        # embedding_output = self.embeddings(
        #     input_ids=input_ids,
        #     position_ids=position_ids,
        #     token_type_ids=token_type_ids,
        #     inputs_embeds=inputs_embeds,
        #     past_key_values_length=past_key_values_length,
        # )
        #
        # encoder_outputs = self.encoder(
        #     embedding_output,
        #     attention_mask=extended_attention_mask,
        #     head_mask=head_mask,
        #     encoder_hidden_states=encoder_hidden_states,
        #     encoder_attention_mask=encoder_extended_attention_mask,
        #     past_key_values=past_key_values,
        #     use_cache=use_cache,
        #     output_attentions=output_attentions,
        #     output_hidden_states=output_hidden_states,
        #     return_dict=return_dict,
        # )
        # print(encoder_outputs)
        sequence_output = input_ids
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        # if not return_dict:
        #     return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=None,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )


# if __name__ == '__main__':
#     m = ImageFeaturesBert(visual_embedding_dim=60)
#     x = torch.ones((2, 10, 60))
#     print(m(x))



import torch
from torch.utils.data import Dataset
from .Seq2Seq import Seq2Seq
from .base.ImageDataset import ImageDataset


class ImSeq2Seq(Dataset):
    def __init__(self, src, tgt, image, split, ckpt_dir, **kwargs):
        self.split = split
        self.seq2seq = Seq2Seq(src, tgt, split, ckpt_dir)
        self.image = ImageDataset(**image, split=split)

        # For decoding
        self.tgt_tokenizer = self.seq2seq.tgt.tokenizer
        self.tgt_tokenizer_max_len = self.seq2seq.tgt.tokenizer_max_len

        assert len(self.image) == len(self.seq2seq)

    def __getitem__(self, index):
        return {**self.image.__getitem__(index), **self.seq2seq.__getitem__(index)}

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {**self.seq2seq.get_collate_fn()(batch), **self.image.get_collate_fn()(batch)}
            return collated

        return collate_fn

    def __len__(self):
        return len(self.seq2seq)

    def __repr__(self):
        return "ImSeq2Seq\n" + str(self.seq2seq) + '\n' + str(self.image)



import tqdm
import torch
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
import functools
from .beam_search import beam_search


def evaluation(models, config, dl, **kwargs):
    hf_models = [model.enc_dec.enc_dec for model in models]

    hf_models[0].beam_search = functools.partial(beam_search, hf_models[0])

    tokenizer = dl.dataset.tgt_tokenizer
    max_len = dl.dataset.tgt_tokenizer_max_len
    ref_list = []
    hyp_list = []

    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            batch = {k: v.cuda() for k, v in batch.items()}
            encoder_batch = {k: v.cuda() for k, v in batch.items() if not k.startswith("decoder_")}

            expanded_return_idx = (
                torch.arange(batch['input_ids'].shape[0]).view(-1, 1).repeat(1, config.beam_width).view(-1).cuda()
            )

            # Run encoder ourselves
            encoders_outputs = [hf.encoder(**encoder_batch, return_dict=True) for hf in hf_models]

            # Tile representations
            expanded_model_outputs = [BaseModelOutputWithPoolingAndCrossAttentions(
                last_hidden_state=h.last_hidden_state.index_select(0, expanded_return_idx)) for h in
                encoders_outputs]
            attention_masks = [batch["attention_mask"].index_select(0, expanded_return_idx) for _ in hf_models]

            # Create model kwargs
            model_kwargs = {"encoder_outputs": encoders_outputs[0],  # Dummy encoder output
                            "model_kwargs_list": [{"encoder_outputs": e,
                                                   "attention_mask": m,
                                                   } for e, m in zip(expanded_model_outputs, attention_masks)
                                                  ],
                            "hf_models": hf_models
                            }

            # Lets go
            hyps = hf_models[0].generate(input_ids=batch["input_ids"],
                                         num_return_sequences=1,
                                         max_length=max_len,
                                         num_beams=config.beam_width,
                                         length_penalty=config.length_penalty,
                                         bos_token_id=hf_models[0].decoder.config.bos_token_id,
                                         eos_token_id=hf_models[0].decoder.config.eos_token_id,
                                         pad_token_id=hf_models[0].decoder.config.pad_token_id,
                                         **model_kwargs
                                         )

            refs = batch["decoder_input_ids"]
            for h, r in zip(hyps, refs):
                hyp_list.append(tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

        return {'refs': ref_list, 'hyps': hyp_list}



import torch.nn as nn
import copy
import functools

# v4.3.2
from transformers.models.auto import AutoModelForCausalLM, AutoConfig
from transformers.models.bert_generation import BertGenerationConfig, BertGenerationDecoder
from vilmedic.networks.blocks.huggingface.decoder.beam_search import prepare_inputs_for_generation
from transformers.models.roberta import modeling_roberta

class DecoderModel(nn.Module):
    """
    If proto is mentioned in decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationDecoder model from decoder dict.
    """

    def __init__(self, decoder, **kwargs):
        super().__init__()
        if decoder.proto is not None:
            path = decoder.pop('proto')
            dec_config = AutoConfig.from_pretrained(path)
            dec_config.is_decoder = True
            dec_config.add_cross_attention = True
            self.decoder = AutoModelForCausalLM.from_pretrained(path, config=dec_config)
        else:
            dec_config = BertGenerationConfig(**decoder)
            dec_config.is_decoder = True
            dec_config.add_cross_attention = True
            self.decoder = BertGenerationDecoder(dec_config)

        # Evaluation
        self.decoder.prepare_inputs_for_generation = functools.partial(prepare_inputs_for_generation, self.decoder)

        # Inference
        self.generate = self.decoder.generate
        self.config = self.decoder.config

    def forward(self, input_ids, attention_mask, encoder_outputs=None, encoder_attention_mask=None, **kwargs):
        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()
        out = self.decoder(input_ids=input_ids,
                           attention_mask=attention_mask,
                           encoder_hidden_states=encoder_outputs,
                           encoder_attention_mask=encoder_attention_mask,
                           labels=input_ids,
                           **kwargs)
        out = vars(out)
        return out

    def __repr__(self):
        s = str(type(self.decoder).__name__) + '(' + str(self.decoder.config) + ')\n'
        return s



import os
import pydicom
import numpy as np
import json
import PIL
import skimage
import logging

import torchxrayvision as xrv
from PIL import Image, ImageFile
from torch.utils.data import Dataset
from torchvision.transforms import *
from .utils import load_file
from .papers.open_image import *

ImageFile.LOAD_TRUNCATED_IMAGES = True  # Are we sure ?
PIL.Image.MAX_IMAGE_PIXELS = None  # Are we sure ?
logging.getLogger('PIL').setLevel(logging.WARNING)


def get_transforms(split, resize, crop, custom_transform_train, custom_transform_val, ext):
    if ext in [".npy", ".npz"]:
        return lambda x: x

    assert not (custom_transform_train is not None and (ext in ['.xrv'])), \
        'Cant specify custom transform when using xrv'

    if ext == '.xrv':
        return transforms.Compose([xrv.datasets.XRayCenterCrop(),
                                   xrv.datasets.XRayResizer(crop),
                                   lambda x: torch.from_numpy(x)])

    if split == 'train':
        return transforms.Compose([
            transforms.Resize(resize),
            transforms.RandomCrop(crop),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406),
                                 (0.229, 0.224, 0.225))]) if custom_transform_train is None else eval(
            custom_transform_train)
    else:
        return transforms.Compose([
            transforms.Resize((crop, crop)),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406),
                                 (0.229, 0.224, 0.225))]) if custom_transform_val is None else eval(
            custom_transform_val)


def open_image(image, ext):
    if ext == '.jpg':
        return Image.open(image).convert('RGB')

    if ext == '.xrv':
        img = skimage.io.imread(image)
        img = xrv.datasets.normalize(img, 255)
        if len(img.shape) > 2:
            img = img[:, :, 0]
        if len(img.shape) < 2:
            print("error, dimension lower than 2 for image")
        return img[None, :, :]

    if ext == '.png':
        return Image.open(image).convert('RGB')

    if ext == '.dcm':
        ds = pydicom.dcmread(image)
        img = ds.pixel_array
        return Image.fromarray(np.uint8(img)).convert('RGB')

    if ext in ['.npy', '.npz']:
        return torch.from_numpy(np.load(image))

    # Special cases
    if ext in PAPER_EXT.keys():
        return eval(PAPER_EXT[ext])(image)

    raise NotImplementedError("Image extension {} not implemented".format(ext))


def make_images(root, image_path, split, file):
    images = load_file(os.path.join(root, split + '.' + file))
    return [os.path.join(image_path, image) for image in images]


class ImageDataset(Dataset):
    def __init__(self,
                 root=None,
                 file=None,
                 split=None,
                 image_path=None,
                 load_memory=False,
                 custom_transform_train=None,
                 custom_transform_val=None,
                 resize=256,
                 crop=224,
                 ext='.jpg',
                 **kwargs):

        assert split is not None, "Argument split cant be None"

        self.root = root
        self.file = file
        self.split = split
        self.image_path = image_path
        self.load_memory = load_memory
        self.resize = eval(str(resize))
        self.crop = eval(str(crop))
        self.ext = ext
        self.images = None

        if file is not None:
            self.images = make_images(root, image_path, split, file)

        self.transform = get_transforms(split,
                                        self.resize,
                                        self.crop,
                                        custom_transform_train,
                                        custom_transform_val,
                                        self.ext)

        if self.load_memory and self.file is not None:
            self.images = [open_image(image, ext) for image in self.images]

    def __len__(self):
        return len(self.images or [])

    def __getitem__(self, index):
        image = self.images[index]
        if not self.load_memory:
            image = open_image(image, self.ext)
        return {'image': self.transform(image)}

    def inference(self, image):
        if not isinstance(image, list):
            image = [image]
        batch = [{'image': self.transform(open_image(i, self.ext))} for i in image]
        return self.get_collate_fn()(batch)

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {'images': torch.stack([s['image'] for s in batch])}
            return collated

        return collate_fn

    def __repr__(self):
        transform = self.transform
        if hasattr(self.transform, "transforms"):
            transform = self.transform.transforms

        return "ImageDataset\n" + \
               json.dumps({
                   "image_path": self.image_path,
                   "root": self.root,
                   "file": self.file,
                   "transforms": transform,
                   "ext": self.ext,
               }, indent=4, sort_keys=False, default=str)



import torch.nn.functional as F
import torch
from einops import rearrange

def prob_mask_like(shape, prob, device):
    return torch.zeros(shape, device = device).float().uniform_(0, 1) < prob

def is_empty(t):
    return t.nelement() == 0

def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d

def forward(
        self,
        text,
        image=None,
        return_loss=False,
        null_cond_prob=0.,
        cache=None,
):
    assert text.shape[
               -1] == self.text_seq_len, f'the length {text.shape[-1]} of the text tokens you passed in does not have the correct length ({self.text_seq_len})'
    batch, device, total_seq_len = text.shape[0], text.device, self.total_seq_len

    # randomly remove text condition with <null_cond_prob> probability

    if null_cond_prob > 0:
        null_mask = prob_mask_like((batch,), null_cond_prob, device=device)
        text *= rearrange(~null_mask, 'b -> b 1')

    # make sure padding in text tokens get unique padding token id

    text_range = torch.arange(self.text_seq_len, device=device) + (self.num_text_tokens - self.text_seq_len)
    text = torch.where(text == 0, text_range, text)

    # add <bos>

    text = F.pad(text, (1, 0), value=0)

    tokens = self.text_emb(text)
    tokens += self.text_pos_emb(torch.arange(text.shape[1], device=device))

    seq_len = tokens.shape[1]

    if exists(image) and not is_empty(image):
        is_raw_image = len(image.shape) == 4

        if is_raw_image:
            image_size = self.vae.image_size
            assert tuple(image.shape[1:]) == (
            3, image_size, image_size), f'invalid image of dimensions {image.shape} passed in during training'

            image = self.vae.get_codebook_indices(image)

        image_len = image.shape[1]
        image_emb = self.image_emb(image)

        image_emb += self.image_pos_emb(image_emb)

        tokens = torch.cat((tokens, image_emb), dim=1)

        seq_len += image_len

    # when training, if the length exceeds the total text + image length
    # remove the last token, since it needs not to be trained

    if tokens.shape[1] > total_seq_len:
        seq_len -= 1
        tokens = tokens[:, :-1]

    if self.stable:
        alpha = 0.1
        tokens = tokens * alpha + tokens.detach() * (1 - alpha)

    if exists(cache) and cache.get('offset'):
        tokens = tokens[:, -1:]
    out = self.transformer(tokens, cache=cache)

    if self.stable:
        out = self.norm_by_max(out)

    logits = self.to_logits(out)

    # mask logits to make sure text predicts text (except last token), and image predicts image

    logits_mask = self.logits_mask[:, :seq_len]
    if exists(cache) and cache.get('offset'):
        logits_mask = logits_mask[:, -1:]
    max_neg_value = -torch.finfo(logits.dtype).max
    logits.masked_fill_(logits_mask, max_neg_value)

    if exists(cache):
        cache['offset'] = cache.get('offset', 0) + logits.shape[1]

    if not return_loss:
        return logits, text, image

    assert exists(image), 'when training, image must be supplied'

    offsetted_image = image + self.num_text_tokens
    labels = torch.cat((text[:, 1:], offsetted_image), dim=1)

    logits = rearrange(logits, 'b n c -> b c n')

    loss_text = F.cross_entropy(logits[:, :, :self.text_seq_len], labels[:, :self.text_seq_len])
    loss_img = F.cross_entropy(logits[:, :, self.text_seq_len:], labels[:, self.text_seq_len:])

    loss = (loss_text + self.loss_img_weight * loss_img) / (self.loss_img_weight + 1)
    return loss


# from . import seq2seq
# from .beam_helpers import generate
from ..beam_helpers import generate
import tqdm
import torch


def beam_search(models, config, dl):
    model = models[0]  # get model attributes
    encdecs = [model.enc_dec for model in models]
    dummy = encdecs[0]
    tgt_tokenizer = dl.dataset.tgt_tokenizer
    ref_list = []
    hyp_list = []
    with torch.no_grad():
        for batch in tqdm.tqdm(dl):
            refs = batch['decoder_input_ids']
            # out = model(**batch)
            # logits = out['logits']
            # hyps = torch.argmax(logits, dim=-1)

            input_ids = batch['input_ids'].cuda()
            input_ids, _ = model.backbone(input_ids)
            input_ids = model.visual_projection(input_ids)
            hyps = generate(dummy,
                            encdecs,
                            input_ids=input_ids,
                            num_return_sequences=1,
                            max_length=dl.dataset.tgt_len,
                            num_beams=config.beam_width,
                            repetition_penalty=config.repetition_penalty,
                            # early_stopping=True,
                            bos_token_id=model.bos_token_id,
                            eos_token_id=model.eos_token_id,
                            pad_token_id=model.pad_token_id,
           )

            for h, r in zip(hyps, refs):
                hyp_list.append(tgt_tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
                ref_list.append(tgt_tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))

        return 0.0, ref_list, hyp_list



from ..vision.cnn import CNN
import torch.nn as nn
from transformers.models.bert.modeling_bert import BertEncoder
from transformers.models.bert_generation import BertGenerationConfig


class CNNTransformer(nn.Module):
    def __init__(self, cnn, transformer, **kwargs):
        super().__init__()
        self.cnn = CNN(**cnn)
        transformer = BertGenerationConfig(transformer)
        self.transformer = BertEncoder(transformer)

    def forward(self):
        pass


import torch.nn as nn
import torch
from collections import OrderedDict
# v4.3.2
from vilmedic.networks.blocks.huggingface.encoder_decoder.evaluation import evaluation
from vilmedic.networks.blocks.huggingface.encoder_decoder.encoder_decoder_model import EncoderDecoderModel


def create_state_dict(ckpt, diff, replace):
    weights = torch.load(ckpt)["model"]
    new_dict = set()
    for k, v in weights.items():
        if diff in k:  # avoid cnn
            new_dict.add((k.replace(diff, replace), v))
    new_dict = OrderedDict(new_dict)
    return new_dict


class RRS(nn.Module):

    def __init__(self, encoder, decoder, **kwargs):
        super().__init__()
        self.enc_dec = EncoderDecoderModel(encoder, decoder)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask):
        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           )
        return out

    def __repr__(self):
        return "SumHugMono\n" + str(self.enc_dec)



import torch.nn as nn
from ..rnn.utils import get_n_params, set_embeddings
import copy

# v4.3.2
import torch
from transformers.models.bert_generation import BertGenerationEncoder, BertGenerationConfig, BertGenerationDecoder
from .beam import beam_search
from transformers.models.roberta import modeling_roberta
from transformers.models.roberta.modeling_roberta import RobertaModel
from transformers.models.encoder_decoder.modeling_encoder_decoder import EncoderDecoderModel
from transformers.modeling_outputs import Seq2SeqLMOutput, BaseModelOutputWithPoolingAndCrossAttentions
from ..vision.cnn import CNN


class MultimodalEnc(nn.Module):
    def __init__(self, encoder, cnn):
        super().__init__()
        self.encoder = encoder
        self.cnn = CNN(**cnn)
        self.visual_projection = nn.Linear(cnn.pop("visual_embedding_dim"), self.encoder.config.hidden_size)

    def forward(
            self,
            input_ids,
            images,
            attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            past_key_values=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
    ):
        output_attentions = output_attentions if output_attentions is not None else self.encoder.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.encoder.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.encoder.config.use_return_dict

        if self.encoder.config.is_decoder:
            use_cache = use_cache if use_cache is not None else self.encoder.config.use_cache
        else:
            use_cache = False

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            batch_size, seq_length = input_shape
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
            batch_size, seq_length = input_shape
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        # past_key_values_length
        past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0

        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length + past_key_values_length)), device=device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.encoder.get_head_mask(head_mask, self.encoder.config.num_hidden_layers)

        with torch.no_grad():
            visual_features, _ = self.cnn(images.cuda())

        # Add visual attribute to encoder_outputs
        visual_features = self.visual_projection(visual_features)

        embedding_output = self.encoder.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
        )

        embedding_output = torch.cat((embedding_output, visual_features), dim=1)

        # update mask accordingly
        image_mask = torch.ones(
            visual_features.size(-2), device=visual_features.device
        ).expand(visual_features.size()[:-1]).long()
        attention_mask = torch.cat((attention_mask, image_mask), dim=-1)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.encoder.get_extended_attention_mask(attention_mask, input_shape,
                                                                                         device)
        encoder_extended_attention_mask = None

        encoder_outputs = self.encoder.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.encoder.pooler(sequence_output) if self.encoder.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        out = BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )
        out.attention_mask = attention_mask
        return out

    def train(self, mode: bool = True):
        self.training = mode
        for module in self.children():
            module.train(mode)
        self.cnn.train(False)
        return self


class MultimodalEncDec(EncoderDecoderModel):
    def __init__(self, encoder, decoder, cnn, **kwargs):
        enc_dec = EncoderDecoderModel.from_encoder_decoder_pretrained(encoder.pop('proto'),
                                                                      decoder.pop('proto'))

        super().__init__(encoder=enc_dec.encoder, decoder=enc_dec.decoder, config=enc_dec.config)
        self.encoder = MultimodalEnc(self.encoder, cnn)
        self.to_tile = ["last_hidden_state", "attention_mask"]

    def forward(
            self,
            input_ids=None,
            attention_mask=None,
            decoder_input_ids=None,
            decoder_attention_mask=None,
            encoder_outputs=None,
            past_key_values=None,
            inputs_embeds=None,
            decoder_inputs_embeds=None,
            labels=None,
            use_cache=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=None,
            **kwargs,
    ):
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            kwargs_encoder = {argument: value for argument, value in kwargs.items() if
                              not argument.startswith("decoder_")}

            # Encoder
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                **kwargs_encoder,
            )

        encoder_hidden_states = encoder_outputs.last_hidden_state
        attention_mask = encoder_outputs.attention_mask

        # Decode
        kwargs_decoder = {
            argument[len("decoder_"):]: value for argument, value in kwargs.items() if argument.startswith("decoder_")
        }
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=attention_mask,
            inputs_embeds=decoder_inputs_embeds,
            labels=labels,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            use_cache=use_cache,
            past_key_values=past_key_values,
            return_dict=return_dict,
            **kwargs_decoder,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqLMOutput(
            loss=decoder_outputs.loss,
            logits=decoder_outputs.logits,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )


class ConcatVisualSeq2SeqHug(nn.Module):
    """
    If proto is mentioned in encoder and decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder/BertGenerationDecoder model from encoder and decoder dict.
    """

    def __init__(self, encoder, decoder, **kwargs):
        super().__init__()

        self.enc_dec = MultimodalEncDec(encoder, decoder, **kwargs)
        self.enc = self.enc_dec.encoder
        self.dec = self.enc_dec.decoder

        self.bos_token_id = self.dec.config.bos_token_id
        self.eos_token_id = self.dec.config.eos_token_id
        self.pad_token_id = self.dec.config.pad_token_id

        # Evaluation
        self.eval_func = beam_search

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask, **kwargs):
        input_ids = input_ids.cuda()
        decoder_input_ids = decoder_input_ids.cuda()
        attention_mask = attention_mask.cuda()
        decoder_attention_mask = decoder_attention_mask.cuda()

        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           labels=decoder_input_ids,
                           **kwargs)
        out = vars(out)
        return out

    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch.nn as nn
from vilmedic.networks.models.utils import get_n_params
import functools
# v4.3.2

from vilmedic.networks.blocks.vision import *
from vilmedic.networks.blocks.huggingface.decoder.evaluation import evaluation as evaluation_

from .RRG import RRG
import numpy as np
import copy
import torch
import logging
import torch.nn.functional as F
import inspect
import random
from vilmedic.scorers.NLG import ROUGEScorer
from vilmedic.scorers.scores import CheXbert


def evaluation(models, config, dl, **kwargs):
    models = [m.model for m in models]  # Get trained RRG instance
    return evaluation_(models, config, dl)


def to_contiguous(tensor):
    if tensor.is_contiguous():
        return tensor
    else:
        return tensor.contiguous()


def scst_loss(input, seq, reward, pad_token_id):
    # HuggingFace TopKLogitsWarper (if top_k > 0) puts -float('inf') for non top_k (or if we use bad_words_ids)
    # Padding can then have logits -float('inf') though we have to pad because hyp generation is finished
    # -float('inf') * 0 masking results in NaN so that doesnt mitigate the issue. need to do:
    input[input == -float("Inf")] = 0.
    #####

    # reward
    N, L = input.shape[:2]
    reward = reward.view(N, 1, 1).expand_as(input)
    input = -input * reward

    # masking
    mask = (seq > pad_token_id).float()
    output = input.view(-1) * mask.view(-1)

    # mean
    output = torch.sum(output) / torch.sum(mask)
    return output


class RRG_SCST(nn.Module):

    def __init__(self, decoder, cnn, ckpt, dl, logger=None, score="chexbert", top_k=None, **kwargs):
        super().__init__()

        # Models
        state_dict = torch.load(ckpt)["model"]
        self.model = RRG(copy.deepcopy(decoder), copy.deepcopy(cnn), **kwargs)
        self.model.load_state_dict(state_dict, strict=True)

        self.top_k = top_k
        self.score = score
        # Scoring
        if score == "chexbert":
            self.chexbert = CheXbert()

        # Tokens
        self.bos_token_id = self.model.dec.decoder.config.bos_token_id
        self.eos_token_id = self.model.dec.decoder.config.eos_token_id
        self.pad_token_id = self.model.dec.decoder.config.pad_token_id

        # Tokenizer
        self.dl = dl
        self.tokenizer = dl.dataset.tokenizer

        self.eval_func = evaluation
        self.logger = logger or logging.get_logger(__name__)

    def forward(self, input_ids, attention_mask, images, encoder_outputs=None, **kwargs):
        batch_size = input_ids.shape[0]
        seq_len = input_ids.shape[1]

        # 1 Greedy
        with torch.no_grad():
            self.model.eval()
            encoder_hidden_states = self.model.encode(images.cuda())
            out = self.model.dec.decoder.generate(
                input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
                max_length=seq_len,
                num_beams=1,
                num_return_sequences=1,
                return_dict_in_generate=True,
                output_scores=True,
                encoder_hidden_states=encoder_hidden_states.data,
                forced_eos_token_id=True,
            )
            greedy_input_ids = out.sequences
            score_greedy = self.get_reward(greedy_input_ids.data, input_ids)

        # 2. Sampling
        self.model.eval()
        # encoder_hidden_states = self.model.encoder(images.cuda())
        out = inspect.unwrap(self.model.dec.decoder.generate)(  # inspect.unwrap removes the torch.no_grad() decorator
            self=self.model.dec.decoder,
            input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
            max_length=seq_len,
            num_beams=1,
            num_return_sequences=1,
            encoder_hidden_states=encoder_hidden_states.data,
            bad_words_ids=[[self.pad_token_id], [self.bos_token_id]],
            top_k=self.top_k,
            forced_eos_token_id=True,
            output_scores=True,
            do_sample=True,
            use_cache=True,
            return_dict_in_generate=True,
        )
        samples_ids = out.sequences[:, 1:]
        logits = torch.stack(out.scores, dim=1)
        logits = F.log_softmax(logits, dim=-1)
        sampled_logits = logits.gather(2, samples_ids.unsqueeze(-1))

        # 3. Reward and loss
        score_sampling = self.get_reward(samples_ids.data, input_ids)
        reward = torch.tensor(score_sampling).cuda() - torch.tensor(score_greedy).cuda()
        loss = scst_loss(sampled_logits, samples_ids.data, reward, self.pad_token_id)

        if random.randint(0, 10) == 0:
            self.logger.info(greedy_input_ids[0])
            self.logger.info(samples_ids[0])
            self.logger.info(score_greedy)
            self.logger.info(score_sampling)
            self.logger.info(loss)
            self.logger.info("#######")

        return {"loss": loss}

    def get_reward(self, rollout_input_ids, input_ids):

        hyp_list = []
        ref_list = []
        for h, r in zip(rollout_input_ids, input_ids):
            hyp_list.append(self.tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
            ref_list.append(self.tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))
        if self.score == "chexbert":
            assert not self.chexbert.training
            chexbert, _ = self.chexbert(hyp_list, ref_list)
            reward = torch.tensor(chexbert["micro avg"]["f1-score"]).cuda()
        else:
            reward = ROUGEScorer(rouges=[self.score]).compute(ref_list, hyp_list)[1]
        return reward

    def __repr__(self):
        s = "RRG_PPO\n"
        # s += str(self.enc) + '\n'
        # s += str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



# -*- coding: utf-8 -*-
import torch.nn.functional as F
from .decoder import ConditionalDecoder
from .layers.attention import Attention
from .layers.hierarchical_attention import HierarchicalAttention


def get_rnn_hidden_state(h):
    """Returns h_t transparently regardless of RNN type."""
    return h if not isinstance(h, tuple) else h[0]


class VisualConditionalDecoder(ConditionalDecoder):

    def __init__(self, visual_ctx_size, **kwargs):
        super().__init__(**kwargs)

        self.visual_ctx_size = visual_ctx_size

        # Create attention layer
        self.visual_att = Attention(self.visual_ctx_size, self.hidden_size,
                                    transform_ctx=self.transform_ctx,
                                    mlp_bias=self.mlp_bias,
                                    att_type=self.att_type,
                                    att_activ=self.att_activ,
                                    att_bottleneck=self.att_bottleneck,
                                    temp=self.att_temp)

        self.fusion = HierarchicalAttention([self.hidden_size, self.hidden_size],
                                            self.hidden_size, self.hidden_size)

    def f_next(self, ctx_dict, y, h):
        h1_c1 = self.dec0(y, self._rnn_unpack_states(h))
        h1 = get_rnn_hidden_state(h1_c1)

        self.txt_alpha_t, txt_z_t = self.att(
            h1.unsqueeze(0), *ctx_dict['enc'])

        self.img_alpha_t, img_z_t = self.visual_att(
            h1.unsqueeze(0), *ctx_dict['feats'])

        self.h_att, z_t = self.fusion([txt_z_t, img_z_t], h1.unsqueeze(0))

        h2_c2 = self.dec1(z_t, h1_c1)
        h2 = get_rnn_hidden_state(h2_c2)

        logit = self.hid2out(h2)

        if self.dropout_out > 0:
            logit = self.do_out(logit)

        log_p = F.log_softmax(self.out2prob(logit), dim=-1)

        return log_p, self._rnn_pack_states(h2_c2)



import subprocess
import os

class BLEUScorer:
    def __init__(self):
        BLEU_SCRIPT = os.path.join(os.path.dirname(__file__), "multi-bleu.perl")
        self.cmdline = [BLEU_SCRIPT]

    def compute(self, refs, hyps):
        refs = [refs] if not isinstance(refs, list) else refs
        self.cmdline.extend(refs)

        if isinstance(hyps, str):
            hypstring = open(hyps).read().strip()
        elif isinstance(hyps, list):
            hypstring = "\n".join(hyps)

        score = subprocess.run(self.cmdline,
                               stdout=subprocess.PIPE,
                               input=hypstring,
                               universal_newlines=True).stdout.splitlines()
        if len(score) == 0:
            return 0.0
        else:
            score = score[0].strip()
            float_score = float(score.split()[2][:-1])
            return float_score



from rouge_score import rouge_scorer
from six.moves import zip_longest
import numpy as np
from rouge_score import scoring


class ROUGEScorer:
    def __init__(self, rouges):
        self.scorer = rouge_scorer.RougeScorer(rouges, use_stemmer=True)
        self.rouges = rouges

    def compute(self, refs, hyps):
        scores = []
        for target_rec, prediction_rec in zip_longest(refs, hyps):
            if target_rec is None or prediction_rec is None:
                raise ValueError("Must have equal number of lines across target and "
                                 "prediction.")
            scores.append(self.scorer.score(target_rec, prediction_rec))

        # aggregator = scoring.BootstrapAggregator()
        # for score in scores:
        #     aggregator.add_scores(score)
        # print(aggregator.aggregate())
        f1_rouge = [s[self.rouges[0]].fmeasure for s in scores]
        return np.mean(f1_rouge), f1_rouge



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00-core.ipynb (unless otherwise specified).

__all__ = ['flatten_dict', 'stack_dicts', 'add_suffix', 'pad_to_size', 'logprobs_from_logits', 'whiten',
           'clip_by_value', 'entropy_from_logits', 'average_torch_dicts', 'stats_to_np', 'build_bert_batch_from_txt']

# Cell
import torch
import torch.nn.functional as F
import collections
import numpy as np

# Cell

def flatten_dict(nested, sep='/'):
    """Flatten dictionary and concatenate nested keys with separator."""
    def rec(nest, prefix, into):
        for k, v in nest.items():
            if sep in k:
                raise ValueError(f"separator '{sep}' not allowed to be in key '{k}'")
            if isinstance(v, collections.Mapping):
                rec(v, prefix + k + sep, into)
            else:
                into[prefix + k] = v
    flat = {}
    rec(nested, '', flat)
    return flat

def stack_dicts(stats_dicts):
    """Stack the values of a dict."""
    results = dict()
    for k in stats_dicts[0]:
        stats_list = [torch.flatten(d[k]) for d in stats_dicts]
        results[k] = torch.stack(stats_list)
    return results

def add_suffix(input_dict, suffix):
    """Add suffix to dict keys."""
    return dict((k + suffix, v) for k,v in input_dict.items())

# Cell

def pad_to_size(tensor, size, dim=1, padding=50256):
    """Pad tensor to size."""
    t_size = tensor.size()[dim]
    if t_size==size:
        return tensor
    else:
        return torch.nn.functional.pad(tensor, (0,size-t_size), 'constant', padding)

def logprobs_from_logits(logits, labels):
    """
    See: https://github.com/pytorch/pytorch/issues/563#issuecomment-330103591
    """
    logp = F.log_softmax(logits, dim=2)
    logpy = torch.gather(logp, 2, labels.unsqueeze(2)).squeeze(-1)
    return logpy


def whiten(values, shift_mean=True):
    """Whiten values."""
    mean, var = torch.mean(values), torch.var(values)
    whitened = (values - mean) * torch.rsqrt(var + 1e-8)
    if not shift_mean:
        whitened += mean
    return whitened

def clip_by_value(x, tensor_min, tensor_max):
    """
    Tensor extenstion to torch.clamp
    https://github.com/pytorch/pytorch/issues/2793#issuecomment-428784713
    """
    clipped = torch.max(torch.min(x, tensor_max), tensor_min)
    return clipped

def entropy_from_logits(logits):
    """Calculate entropy from logits."""
    pd = torch.nn.functional.softmax(logits, dim=-1)
    entropy = torch.logsumexp(logits, axis=-1) - torch.sum(pd*logits, axis=-1)
    return entropy


def average_torch_dicts(list_of_dicts):
    """Average values of a list of dicts wiht torch tensors."""
    average_dict = dict()
    for key in list_of_dicts[0].keys():
        average_dict[key] = torch.mean(torch.stack([d[key] for d in list_of_dicts]), axis=0)
    return average_dict

def stats_to_np(stats_dict):
    """Cast all torch.tensors in dict to numpy arrays."""
    new_dict = dict()
    for k, v in stats_dict.items():
        if isinstance(v, torch.Tensor):
            new_dict[k] = v.detach().cpu().numpy()
        else:
            new_dict[k] = v
        if np.isscalar(new_dict[k]):
            new_dict[k] = float(new_dict[k])
    return new_dict


# Cell

def build_bert_batch_from_txt(text_list, tokenizer, device):
    """Create token id and attention mask tensors from text list for BERT classification."""

    # tokenize
    tensors = [tokenizer.encode(txt, return_tensors="pt").to(device) for txt in text_list]

    # find max length to pad to
    max_len = max([t.size()[1] for t in tensors])

    # get padded tensors and attention masks
    # (attention masks make bert ignore padding)
    padded_tensors = []
    attention_masks = []
    for tensor in tensors:
        attention_mask = torch.ones(tensor.size(), device=device)
        padded_tensors.append(pad_to_size(tensor, max_len, padding=0))
        attention_masks.append(pad_to_size(attention_mask, max_len, padding=0))

    # stack all tensors
    padded_tensors = torch.cat(padded_tensors)
    attention_masks = torch.cat(attention_masks)

    return padded_tensors, attention_masks


# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

from openvqa.core.path_cfgs import PATH
import os, torch, random
import numpy as np
from types import MethodType


class BaseCfgs(PATH):
    def __init__(self):
        super(BaseCfgs, self).__init__()

        # Set Devices
        # If use multi-gpu training, you can set e.g.'0, 1, 2' instead
        self.GPU = '0'

        # Set Seed For CPU And GPUs
        self.SEED = random.randint(0, 9999999)

        # -------------------------
        # ---- Version Control ----
        # -------------------------

        # You can set a name to start new training
        self.VERSION = str(self.SEED)

        # Use checkpoint to resume training
        self.RESUME = False

        # Resume training version or testing version
        self.CKPT_VERSION = self.VERSION

        # Resume training epoch or testing epoch
        self.CKPT_EPOCH = 0

        # if set 'CKPT_PATH', -> 'CKPT_VERSION' and 'CKPT_EPOCH' will not work any more
        self.CKPT_PATH = None

        # Print loss every iteration
        self.VERBOSE = True


        # ------------------------------
        # ---- Data Provider Params ----
        # ------------------------------

        self.MODEL = ''

        self.MODEL_USE = ''

        self.DATASET = ''

        # Run as 'train' 'val' or 'test'
        self.RUN_MODE = ''

        # Set True to evaluate offline when an epoch finished
        # (only work when train with 'train' split)
        self.EVAL_EVERY_EPOCH = True

        # Set True to save the prediction vector
        # (use in ensemble)
        self.TEST_SAVE_PRED = False


        # A external method to set train split
        # will override the SPLIT['train']
        self.TRAIN_SPLIT = 'train'

        # Set True to use pretrained GloVe word embedding
        # (GloVe: spaCy https://spacy.io/)
        self.USE_GLOVE = True

        # Word embedding matrix size
        # (token size x WORD_EMBED_SIZE)
        self.WORD_EMBED_SIZE = 300

        # All features size
        self.FEAT_SIZE = {
            'vqa': {
                'FRCN_FEAT_SIZE': (100, 2048),
                'BBOX_FEAT_SIZE': (100, 5),
            },
            'gqa': {
                'FRCN_FEAT_SIZE': (100, 2048),
                'GRID_FEAT_SIZE': (49, 2048),
                'BBOX_FEAT_SIZE': (100, 5),
            },
            'clevr': {
                'GRID_FEAT_SIZE': (196, 1024),
            },
        }

        # Set if bbox_feat need be normalize by image size, default: False
        self.BBOX_NORMALIZE = False

        # Default training batch size: 64
        self.BATCH_SIZE = 64

        # Multi-thread I/O
        self.NUM_WORKERS = 8

        # Use pin memory
        # (Warning: pin memory can accelerate GPU loading but may
        # increase the CPU memory usage when NUM_WORKS is big)
        self.PIN_MEM = True

        # Large model can not training with batch size 64
        # Gradient accumulate can split batch to reduce gpu memory usage
        # (Warning: BATCH_SIZE should be divided by GRAD_ACCU_STEPS)
        self.GRAD_ACCU_STEPS = 1


        # --------------------------
        # ---- Optimizer Params ----
        # --------------------------

        # Define the loss function
        '''
        Loss(case-sensitive): 
        'ce'    : Cross Entropy -> NLLLoss(LogSoftmax(output), label) = CrossEntropyLoss(output, label)
        'bce'   : Binary Cross Entropy -> BCELoss(Sigmoid(output), label) = BCEWithLogitsLoss(output, label)
        'kld'   : Kullback-Leibler Divergence -> KLDivLoss(LogSoftmax(output), Softmax(label))
        'mse'   : Mean Squared Error -> MSELoss(output, label)
        
        Reduction(case-sensitive):
        'none': no reduction will be applied
        'elementwise_mean': the sum of the output will be divided by the number of elements in the output
        'sum': the output will be summed
        '''
        self.LOSS_FUNC = ''
        self.LOSS_REDUCTION = ''


        # The base learning rate
        self.LR_BASE = 0.0001

        # Learning rate decay ratio
        self.LR_DECAY_R = 0.2

        # Learning rate decay at {x, y, z...} epoch
        self.LR_DECAY_LIST = [10, 12]

        # Warmup epoch lr*{1/(n+1), 2/(n+1), ... , n/(n+1)}
        self.WARMUP_EPOCH = 3

        # Max training epoch
        self.MAX_EPOCH = 13

        # Gradient clip
        # (default: -1 means not using)
        self.GRAD_NORM_CLIP = -1

        # Optimizer
        '''
        Optimizer(case-sensitive): 
        'Adam'      : default -> {betas:(0.9, 0.999), eps:1e-8, weight_decay:0, amsgrad:False}
        'Adamax'    : default -> {betas:(0.9, 0.999), eps:1e-8, weight_decay:0}
        'RMSprop'   : default -> {alpha:0.99, eps:1e-8, weight_decay:0, momentum:0, centered:False}
        'SGD'       : default -> {momentum:0, dampening:0, weight_decay:0, nesterov:False}
        'Adadelta'  : default -> {rho:0.9, eps:1e-6, weight_decay:0}
        'Adagrad'   : default -> {lr_decay:0, weight_decay:0, initial_accumulator_value:0}
        
        In YML files:
        If you want to self-define the optimizer parameters, set a dict named OPT_PARAMS contains the keys you want to modify.
         !!! Warning: keys: ['params, 'lr'] should not be set. 
         !!! Warning: To avoid ambiguity, the value of keys should be defined as string type.
        If you not define the OPT_PARAMS, all parameters of optimizer will be set as default.
        Example:
        mcan_small.yml ->
            OPT: Adam
            OPT_PARAMS: {betas: '(0.9, 0.98)', eps: '1e-9'}
        '''
        # case-sensitive
        self.OPT = ''
        self.OPT_PARAMS = {}


    def str_to_bool(self, args):
        bool_list = [
            'EVAL_EVERY_EPOCH',
            'TEST_SAVE_PRED',
            'RESUME',
            'PIN_MEM',
            'VERBOSE',
        ]

        for arg in dir(args):
            if arg in bool_list and getattr(args, arg) is not None:
                setattr(args, arg, eval(getattr(args, arg)))

        return args


    def parse_to_dict(self, args):
        args_dict = {}
        for arg in dir(args):
            if not arg.startswith('_') and not isinstance(getattr(args, arg), MethodType):
                if getattr(args, arg) is not None:
                    args_dict[arg] = getattr(args, arg)

        return args_dict


    def add_args(self, args_dict):
        for arg in args_dict:
            setattr(self, arg, args_dict[arg])


    def proc(self):
        assert self.RUN_MODE in ['train', 'val', 'test']

        # ------------ Devices setup
        os.environ['CUDA_VISIBLE_DEVICES'] = self.GPU
        self.N_GPU = len(self.GPU.split(','))
        self.DEVICES = [_ for _ in range(self.N_GPU)]
        torch.set_num_threads(2)


        # ------------ Path check
        self.check_path(self.DATASET)


        # ------------ Model setup (Deprecated)
        # self.MODEL_USE = self.MODEL.split('_')[0]


        # ------------ Seed setup
        # fix pytorch seed
        torch.manual_seed(self.SEED)
        if self.N_GPU < 2:
            torch.cuda.manual_seed(self.SEED)
        else:
            torch.cuda.manual_seed_all(self.SEED)
        torch.backends.cudnn.deterministic = True

        # fix numpy seed
        np.random.seed(self.SEED)

        # fix random seed
        random.seed(self.SEED)

        if self.CKPT_PATH is not None:
            print("Warning: you are now using 'CKPT_PATH' args, "
                  "'CKPT_VERSION' and 'CKPT_EPOCH' will not work")
            self.CKPT_VERSION = self.CKPT_PATH.split('/')[-1] + '_' + str(random.randint(0, 9999999))


        # ------------ Split setup
        self.SPLIT = self.SPLITS[self.DATASET]
        self.SPLIT['train'] = self.TRAIN_SPLIT
        if self.SPLIT['val'] in self.SPLIT['train'].split('+') or self.RUN_MODE not in ['train']:
            self.EVAL_EVERY_EPOCH = False

        if self.RUN_MODE not in ['test']:
            self.TEST_SAVE_PRED = False


        # ------------ Gradient accumulate setup
        assert self.BATCH_SIZE % self.GRAD_ACCU_STEPS == 0
        self.SUB_BATCH_SIZE = int(self.BATCH_SIZE / self.GRAD_ACCU_STEPS)

        # Set small eval batch size will reduce gpu memory usage
        self.EVAL_BATCH_SIZE = int(self.SUB_BATCH_SIZE / 2)


        # ------------ Loss process
        assert self.LOSS_FUNC in ['ce', 'bce', 'kld', 'mse']
        assert self.LOSS_REDUCTION in ['none', 'elementwise_mean', 'sum']

        self.LOSS_FUNC_NAME_DICT = {
            'ce': 'CrossEntropyLoss',
            'bce': 'BCEWithLogitsLoss',
            'kld': 'KLDivLoss',
            'mse': 'MSELoss',
        }

        self.LOSS_FUNC_NONLINEAR = {
            'ce': [None, 'flat'],
            'bce': [None, None],
            'kld': ['log_softmax', None],
            'mse': [None, None],
        }

        self.TASK_LOSS_CHECK = {
            'vqa': ['bce', 'kld'],
            'gqa': ['ce'],
            'clevr': ['ce'],
        }

        assert self.LOSS_FUNC in self.TASK_LOSS_CHECK[self.DATASET], \
            self.DATASET + 'task only support' + str(self.TASK_LOSS_CHECK[self.DATASET]) + 'loss.' + \
            'Modify the LOSS_FUNC in configs to get a better score.'


        # ------------ Optimizer parameters process
        assert self.OPT in ['Adam', 'Adamax', 'RMSprop', 'SGD', 'Adadelta', 'Adagrad']
        optim = getattr(torch.optim, self.OPT)
        default_params_dict = dict(zip(optim.__init__.__code__.co_varnames[3: optim.__init__.__code__.co_argcount],
                                       optim.__init__.__defaults__[1:]))

        def all(iterable):
            for element in iterable:
                if not element:
                    return False
            return True
        assert all(list(map(lambda x: x in default_params_dict, self.OPT_PARAMS)))

        for key in self.OPT_PARAMS:
            if isinstance(self.OPT_PARAMS[key], str):
                self.OPT_PARAMS[key] = eval(self.OPT_PARAMS[key])
            else:
                print("To avoid ambiguity, set the value of 'OPT_PARAMS' to string type")
                exit(-1)
        self.OPT_PARAMS = {**default_params_dict, **self.OPT_PARAMS}

    def __str__(self):
        __C_str = ''
        for attr in dir(self):
            if not attr.startswith('__') and not isinstance(getattr(self, attr), MethodType):
                __C_str += '{ %-17s }->' % attr + str(getattr(self, attr)) + '\n'

        return __C_str


#
#
# if __name__ == '__main__':
#     __C = Cfgs()
#     __C.proc()








from radgraph.inference_from_script import get_rewards

report = "FINAL REPORT INDICATION : ___ F with cough / / Cough TECHNIQUE : PA and lateral views of the chest . COMPARISON : None . FINDINGS : The lungs are clear without focal consolidation , , or edema . The cardiomediastinal silhouette is within normal limits . No acute osseous abnormalities . IMPRESSION : No acute cardiopulmonary process ."
hypothesis_report_list = [report, "", "a", report]

report_2 = "FINAL REPORT INDICATION : ___ F with cough / / Cough TECHNIQUE : PA and lateral views of the chest . COMPARISON : None . FINDINGS : The heart is clear without focal consolidation , , or edema . The cardiomediastinal silhouette is within normal limits . No acute osseous abnormalities . IMPRESSION : No acute cardiopulmonary process ."
reference_report_list = [report_2, report_2, report_2, report_2]

reward_list = get_rewards(hypothesis_report_list, reference_report_list, 0.5,
                          0.5, "partial", -1)


print(reward_list)  # [0.8666666666666667, 0, 0, 0.8666666666666667]



import numpy as np
import logging
import torch
import tqdm
import sys
import os

from .validator import Validator
from .utils import CheckpointSaver, create_model, create_data_loader, create_optimizer, create_training_scheduler, \
    create_scaler


class InitTrainor(object):
    def __init__(self, config, seed):
        self.seed = seed
        self.config = config
        self.state = None

        # Do we resume training?
        if config.ckpt is not None:
            self.state = torch.load(config.ckpt)

        # Logger
        self.logger = logging.getLogger(str(seed))

        # Checkpoints
        self.saver = CheckpointSaver(ckpt_dir=self.config.ckpt_dir, logger=self.logger, seed=self.seed,
                                     ckpt=self.config.ckpt)

        # Dataloader
        self.dl = create_data_loader(self.config, split='train', logger=self.logger)

        # Model
        self.model = create_model(self.config, dl=self.dl, logger=self.logger, state_dict=self.state)

        # Optimizer
        self.optimizer = create_optimizer(config=self.config, logger=self.logger, params=self.model.parameters(),
                                          state_dict=self.state)

        # Lr Scheduler and early stop
        self.training_scheduler = create_training_scheduler(config=self.config, optimizer=self.optimizer,
                                                            logger=self.logger, state_dict=self.state)

        # Scaler
        self.scaler = create_scaler(config=self.config, logger=self.logger, state_dict=self.state)

        # Training
        self.eval_start = self.config.eval_start
        self.grad_accu = self.config.grad_accu or 1
        self.clip_grad_norm = self.config.clip_grad_norm

        # Validator is None at init
        self.evaluator: Validator = None


class Trainor(InitTrainor):
    def __init__(self, config, seed):
        super().__init__(config=config, seed=seed)

    def start(self):

        for epoch in range(int(self.training_scheduler.epoch), self.config.epochs + 1):
            self.model.train()

            losses = []
            log = ""
            do_eval = epoch > self.eval_start - 1
            pbar = tqdm.tqdm(self.dl, total=len(self.dl))

            # Training
            for iteration, batch in enumerate(pbar, start=1):
                with torch.cuda.amp.autocast(enabled=self.scaler.is_enabled()):
                    if type(batch) is dict:
                        out = self.model(**batch)
                    else:
                        out = self.model(batch)

                # If the model is taking care of it own training
                if 'loss' not in out:
                    pbar.set_description('Epoch {}, {}'.format(epoch + 1, out))
                    continue

                loss = out['loss']
                if isinstance(self.model, torch.nn.DataParallel):
                    loss = loss.mean()
                self.scaler.scale(loss).backward()

                if self.clip_grad_norm is not None:
                    self.scaler.unscale_(self.optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.config.clip_grad_norm)

                losses.append(loss.item())

                if iteration % self.grad_accu == 0:
                    self.scaler.step(self.optimizer)
                    self.scaler.update()
                    self.optimizer.zero_grad()

                    log = 'Epoch {}, Lr {}, Loss {:.2f}, {} {:.2f}, ES {}'.format(
                        epoch + 1,
                        [param_group['lr'] for param_group in self.optimizer.param_groups],
                        sum(losses) / iteration,
                        self.config.early_stop_metric,
                        self.training_scheduler.current_best_metric,
                        self.training_scheduler.early_stop,
                    )
                    pbar.set_description(log)
                # break

            # Perform last update if needed
            if (iteration % self.grad_accu != 0) and ('loss' in out):
                if self.clip_grad_norm is not None:
                    self.scaler.unscale_(self.optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.config.clip_grad_norm)
                self.scaler.step(self.optimizer)
                self.scaler.update()
                self.optimizer.zero_grad()

            self.logger.info(log)

            # Evaluation
            mean_eval_metric = None
            if do_eval:
                self.evaluator.epoch = epoch
                self.evaluator.start()
                mean_eval_metric = np.mean([s[self.config.early_stop_metric] for s in self.evaluator.scores])

            # End of epochs instructions
            ret = self.training_scheduler.step(mean_eval_metric, training_loss=sum(losses) / iteration)
            if ret["done_training"]:
                self.logger.info("Early stopped reached")
                sys.exit()
            if ret["save_state"]:
                self.saver.save(state_dict={"model": self.model.state_dict(),
                                            "training_scheduler": self.training_scheduler.state_dict(),
                                            "optimizer": self.optimizer.state_dict(),
                                            "config": self.config,
                                            "scaler": self.scaler.state_dict()
                                            },
                                tag=mean_eval_metric,
                                current_epoch=epoch,
                                )



import torch.nn as nn
from ..rnn.utils import get_n_params, set_embeddings
import copy

# v4.3.2
from transformers.models.bert_generation import BertGenerationEncoder, BertGenerationConfig, BertGenerationDecoder
from transformers import EncoderDecoderModel
from .beam import beam_search


class Seq2SeqHug(nn.Module):
    """
    If proto is mentioned in encoder and decoder dict, loads pretrained models from proto strings.
    Otherwise, loads a BertGenerationEncoder/BertGenerationDecoder model from encoder and decoder dict.
    """

    def __init__(self, encoder, decoder, **kwargs):
        super().__init__()
        if 'proto' in decoder and 'proto' in encoder:
            self.enc_dec = EncoderDecoderModel.from_encoder_decoder_pretrained(encoder.pop('proto'),
                                                                               decoder.pop('proto'))

            self.enc = self.enc_dec.encoder
            self.dec = self.enc_dec.decoder
        else:
            # Encoder
            encoder = vars(encoder)
            enc_config = BertGenerationConfig(**encoder)
            self.enc = BertGenerationEncoder(enc_config)

            # Decoder
            decoder = vars(decoder)
            dec_config = copy.deepcopy(enc_config)
            dec_config.update(decoder)
            dec_config.is_decoder = True
            dec_config.add_cross_attention = True
            self.dec = BertGenerationDecoder(dec_config)

            # Encdec
            self.enc_dec = EncoderDecoderModel(encoder=self.enc, decoder=self.dec)

        # Tokens
        self.bos_token_id = self.dec.config.bos_token_id
        self.eos_token_id = self.dec.config.eos_token_id
        self.pad_token_id = self.dec.config.pad_token_id

        # Evaluation
        self.eval_func = beam_search
        self.generate_config = {}

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask):
        input_ids = input_ids.cuda()
        decoder_input_ids = decoder_input_ids.cuda()
        attention_mask = attention_mask.cuda()
        decoder_attention_mask = decoder_attention_mask.cuda()
        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           labels=decoder_input_ids)
        out = vars(out)
        return out

    def __repr__(self):
        # s = super().__repr__() + '\n'
        s = str(type(self.enc_dec.encoder).__name__) + '(' + str(self.enc_dec.encoder.config) + ')\n'
        s += str(type(self.enc_dec.decoder).__name__) + '(' + str(self.enc_dec.decoder.config) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from utils import get_args, get, print_args, get_seed, extract_seed_from_ckpt
from logger import set_logger
from vilmedic.executors import Trainor, Validator


def main():
    # Get args and create seed
    config, override = get_args()
    seed = get_seed()

    # Create checkpoint dir
    config.ckpt_dir = os.path.join(config.ckpt_dir, config.name)
    os.makedirs(config.ckpt_dir, exist_ok=True)

    # If ckpt is specified, we continue training. Lets extract seed
    if config.ckpt is not None:
        config.ckpt = os.path.join(config.ckpt_dir, config.ckpt)
        seed = extract_seed_from_ckpt(config.ckpt)

    # Create logger according to seed
    set_logger(config.ckpt_dir, seed)

    # Nice print args
    print_args(config, ['trainor', 'validator'], seed, override)

    # Fetch args for training and validation
    train_config = get(config, 'trainor')
    val_config = get(config, 'validator')

    # Trainor
    trainor = Trainor(config=train_config,  # train_config is all args but the other executors args
                      seed=seed)

    # Evaluator
    evaluator = Validator(config=val_config,
                          models=[trainor.model],
                          seed=seed,
                          from_training=True)

    # Lets be gentle, give evaluator to trainor
    trainor.evaluator = evaluator

    # Boom
    trainor.start()


if __name__ == "__main__":
    main()



# --------------------------------------------------------
# OpenVQA
# Written by Yuhao Cui https://github.com/cuiyuhao1996
# --------------------------------------------------------

import torch.nn as nn
import torch

class LayerNorm(nn.Module):
    def __init__(self, size, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.eps = eps

        self.a_2 = nn.Parameter(torch.ones(size))
        self.b_2 = nn.Parameter(torch.zeros(size))

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)

        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2
